"""

    ElutionMapperPlotter.py

        recognition of peaks

    Copyright (c) 2018-2024, SAXS Team, KEK-PF

"""
import os
import numpy                as np
import copy
from scipy.interpolate      import UnivariateSpline
import matplotlib
import matplotlib.pyplot    as plt
import matplotlib.patches   as mpl_patches      # 'as patches' does not work properly
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter             import Tk
from molass_legacy.KekLib.OurMatplotlib          import NavigationToolbar
import OurMessageBox        as MessageBox
from molass_legacy.KekLib.TkUtils                import is_low_resolution, split_geometry
from molass_legacy.KekLib.BasicUtils             import get_caller_module
from molass_legacy.KekLib.TkSupplements          import tk_set_icon_portable, BlinkingFrame
from molass_legacy.UV.AbsorbanceViewer import AbsorbanceViewer
from PlotAnnotation         import add_flow_change_annotation
from molass_legacy._MOLASS.SerialSettings         import get_setting, set_setting, get_xray_picking
from DevSettings            import get_dev_setting
from molass_legacy._MOLASS.SerialSettings         import get_setting, set_setting
from molass_legacy.KekLib.ExceptionTracebacker   import ExceptionTracebacker
from molass_legacy.SerialAnalyzer.ElutionCurve import MIN_HALF_WIDTH

DRAW_MAPPING_RANGES         = True
SUGGEST_DISCOMFORT_BY_TEXT  = True
FIGURE_WIDTH_POSITIONING    = 465
VALLEY_BOUNDARY_COLOR = "gray"
from .SingleComponent import SCI_BOUNDARIES

diff_btn_texts = [ 'Hide difference', 'Show difference', 'Show abs(difference)' ]

class ElutionMapperPlotter:
    def __init__( self, dialog, body_frame, sd, mapper=None, anim_mode=False ):

        self.mplt_ge_2_2 = matplotlib.__version__ >= '2.2'
        self.dialog = dialog
        self.logger = dialog.logger
        self.uv_j0 = sd.uv_j0
        self.xr_j0 = sd.xr_j0
        self.developer_mode     = False
        self.mapper             = mapper
        self.set_range_info(mapper)
        self.anim_mode          = anim_mode
        self.absorbance_picking = get_setting( 'absorbance_picking' )
        self.intensity_picking  = get_xray_picking()
        self.using_xray_conc = get_setting('use_xray_conc') == 1

        self.mapping_show_mode  = 'locally'
        self.in_range_adjustment = False    # should be changed depending on self.initial_std_diff
        self.popup_menu = None
        self.last_click_event = None
        uv_size = len(mapper.a_vector)
        self.x1 = self.uv_j0 + np.arange(uv_size)
        xray_size = len(mapper.x_vector)
        self.x2 = self.xr_j0 + np.arange(xray_size)
        self.x3 = self.x2

        self.hiresolution       =  get_dev_setting( 'hiresolution' )
        if is_low_resolution():
            figsize=( 16, 8 ) if self.developer_mode else ( 16, 4.4 )
        else:
            if self.hiresolution:
                figsize=( 23, 8 )
            else:
                figsize=( 18, 9 ) if self.developer_mode else ( 18, 5.5 )

        self.fig = fig = plt.figure( figsize=figsize )
        ax1 = self.fig.add_subplot( 131 )
        ax2 = self.fig.add_subplot( 132 )
        ax3 = self.fig.add_subplot( 133 )
        self.axes = [ ax1, ax2, ax3 ]

        cframe = Tk.Frame( body_frame )
        cframe.pack()
        tframe = Tk.Frame( body_frame )
        tframe.pack(fill=Tk.X)

        if self.hiresolution:
            for ax in self.axes:
                ax.tick_params( labelsize=16 )

        self.mpl_canvas = FigureCanvasTkAgg( self.fig, cframe )
        self.show()
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )
        self.mpl_canvas.mpl_connect('button_press_event', self.on_mpl_button_press)

        btn_frame = Tk.Frame( tframe )
        btn_frame.pack(side=Tk.LEFT, fill=Tk.X, expand=1)
        mpl_tb_frame = Tk.Frame(btn_frame)
        mpl_tb_frame.pack(side=Tk.LEFT, padx=10)
        self.toolbar = NavigationToolbar(self.mpl_canvas, mpl_tb_frame)
        self.toolbar.update()

        other_btn_frame = Tk.Frame(btn_frame)
        other_btn_frame.pack(side=Tk.RIGHT)
        space = Tk.Frame(other_btn_frame, width=70)
        space.pack(side=Tk.RIGHT)

        self.ax3t = None
        self.diff_visible = 0
        self.show_diff_btn = Tk.Button(other_btn_frame, command=self.show_hide_difference)
        self.show_diff_btn.pack(side=Tk.RIGHT, padx=10)
        self.other_btn_frame = other_btn_frame  # for self.dialog.show_button

    def close_fig(self):
        plt.close(self.fig)

    def set_range_info(self, mapper):
        """
        mapper.mapping_ranges
        mapper.uv_peak_eval_ranges
        mapper.peak_eval_ranges
        """

        def shift_ranges(j0, ranges):
            return [ [ j0 + r for r in range_ ] for range_ in ranges ]

        ranges_ = []
        for erec in mapper.x_curve.get_default_editor_ranges():
            if len(erec) == 1:
                # temporary fix for Kosugi3a
                left, right = erec[0]
                range_ = [left, left+MIN_HALF_WIDTH, right]
            else:
                range_ = [*erec[0], erec[1][1]]
            ranges_.append(range_)
        self.target_ranges = ranges_
        self.shifted_target_ranges = shift_ranges( self.xr_j0, self.target_ranges )
        self.uv_peak_eval_ranges = shift_ranges( self.uv_j0, mapper.uv_peak_eval_ranges )
        self.peak_eval_ranges = shift_ranges( self.xr_j0, mapper.peak_eval_ranges )

    def get_target_ranges(self):
        # for Tester
        return self.target_ranges

    def get_figsize( self ):
        return self.fig.get_size_inches()*self.fig.dpi

    def get_canvas_width( self ):
        return int( self.mpl_canvas_widget.cget( 'width' ) )

    def show( self ):
        if self.mplt_ge_2_2:
            self.mpl_canvas.draw()
        else:
            self.mpl_canvas.show()

    def get_draw_params(self):
        mapper = self.mapper
        base_color = self.get_baseline_color(mapper)
        cd_colors = self.dialog.get_cd_colors()
        opt_params = mapper.opt_params
        ub_adjust  = opt_params['uv_baseline_adjust'] == 1
        xb_adjust  = opt_params['xray_baseline_adjust'] == 1
        return mapper, ub_adjust, xb_adjust, base_color, cd_colors

    def draw( self, clear=False, restrict_info=None, animation_counter=None, animation_done=None ):
        ax1, ax2= self.axes[0:2]
        if clear:
            for ax in [ ax1, ax2 ]:
                ax.cla()

        mapper = self.mapper

        self.sci_list = mapper.get_sci_list()

        if self.anim_mode:
            self.anim_counter_text = '  i=(%d, %d)' % ( mapper.opt_phase, animation_counter)
        else:
            self.anim_counter_text = ''

        if self.hiresolution:
            self.title_fonsize   = 16
        else:
            self.title_fonsize   = 14

        in_folder = get_setting('in_folder')
        self.fig.suptitle( 'Mapping status between UV and Xray elutions from ' + in_folder, fontsize=self.title_fonsize )

        self.draw_fig1(ax1)
        self.draw_fig2(ax2)

        self.the_curves1 = [["standard uv elution", self.x1, mapper.a_vector]]
        self.the_curves2 = [["standard xray elution", self.x2, mapper.x_vector]]

        ymin1, ymax1 = ax1.get_ylim()
        ymin2, ymax2 = ax2.get_ylim()

        if restrict_info is None:
            self.draw_mapped(self.axes[2], clear=clear)
        else:
            print( 'draw: restrict_info=', restrict_info )
            for x in restrict_info:
                ax2.plot( [ x, x ], [ ymin2, ymax2 ], color='yellow' )

        if self.anim_mode:
            for ax in self.axes:
                xmin, xmax = ax.get_xlim()
                ymin, ymax = ax.get_ylim()
                tx = xmin*0.8 + xmax*0.2
                ty = ymin*0.5 + ymax*0.5
                text = 'Complete' if animation_done else "Phase %d" % mapper.opt_phase
                ax.text( tx, ty, text, alpha=0.1, fontsize=60 )

        self.show()
        self.fig.tight_layout()
        self.fig.subplots_adjust( top=0.87, right=0.955 )

    def draw_fig1(self, ax, draw_ranges=DRAW_MAPPING_RANGES):
        mapper, ub_adjust, _, base_color, cd_colors = self.get_draw_params()

        uv_source = 'Microfluidic Control Info' if self.mapper.get_conc_type() == 2 else 'UV absorbance'
        ax.set_title( '(a)    Elution from %s at λ=%.4g%s' % ( uv_source, self.absorbance_picking, self.anim_counter_text ), fontsize=self.title_fonsize )

        ax.plot( self.x1, mapper.a_vector, color='blue' )

        if ub_adjust:
            ax.plot( self.x1, mapper.a_base, color='pink' )
            a_base_ = mapper.a_base + mapper.a_base_adjustment
            ax.plot( self.x1, a_base_, color=base_color, alpha=0.5 )
            self.add_adjust_polygon(ax, self.uv_j0, mapper.a_base, a_base_)
        else:
            assert mapper.a_base_adjustment == 0
            ax.plot( self.x1, mapper.a_base, color=base_color )

        for i, info in enumerate(mapper.a_curve.peak_info):
            peak    = info[1]
            ax.plot( self.uv_j0 + peak, mapper.a_spline(peak), 'o', color=cd_colors[i] )

        if mapper.inv_mapped_boundaries is not None:
            for boudary in mapper.inv_mapped_boundaries:
                ax.plot( self.uv_j0 + boudary, mapper.a_spline(boudary), 'o', color=VALLEY_BOUNDARY_COLOR )

        ymin1, ymax1 = ax.get_ylim()
        ax.set_ylim( ymin1, ymax1 )
        if self.mapper.get_conc_type() < 2:
            flow_changes = mapper.flow_changes
            add_flow_change_annotation( ax, flow_changes, mapper )
        else:
            # TODO: self.helper_info seems to contain incorrect info
            pass

        if self.anim_mode:
            anim_color = 'pink' if mapper.in_opt else None
        else:
            anim_color = None
        self.add_range_patches( ax, ymin1, ymax1, self.uv_peak_eval_ranges, color=anim_color )

        if draw_ranges:
            A, B = mapper.map_params
            for k, range_ in enumerate( self.target_ranges ):
                lower, _, upper = range_
                a_lower = A*lower + B
                a_upper = A*upper + B
                for x in [ a_lower, a_upper ]:
                    x_ = self.uv_j0 + x
                    ax.plot( [ x_, x_ ], [ ymin1, ymax1 ], ':', color='black', alpha=0.2 )

    def draw_fig2(self, ax, title=True, draw_ranges=DRAW_MAPPING_RANGES, peak_no=None):
        mapper, _, xb_adjust, base_color, cd_colors = self.get_draw_params()

        if title:
            ax.set_title( '(b)    Elution from Xray scattering around q=%.3g%s' % ( self.intensity_picking, self.anim_counter_text ), fontsize=self.title_fonsize )

        ax.plot( self.x2, mapper.x_vector, color='orange' )

        if xb_adjust:
            ax.plot( self.x2, mapper.x_base, color='pink' )
            x_base_ = mapper.x_base + mapper.x_base_adjustment
            ax.plot( self.x2, x_base_, color=base_color, alpha=0.5 )
            self.add_adjust_polygon(ax, self.xr_j0, mapper.x_base, x_base_)
        else:
            # assert mapper.x_base_adjustment == 0
            ax.plot(self.x2, mapper.x_base, color=base_color)

        for i, info in enumerate( mapper.x_curve.peak_info ):
                peak    = info[1]
                ax.plot(self.xr_j0 + peak, mapper.x_spline(peak), 'o', color=cd_colors[i])

        for boudary in mapper.x_curve.boundaries:
            ax.plot(self.xr_j0 + boudary, mapper.x_spline(boudary), 'o', color=VALLEY_BOUNDARY_COLOR)

        ymin2, ymax2 = ax.get_ylim()
        ax.set_ylim( ymin2, ymax2 )

        if SUGGEST_DISCOMFORT_BY_TEXT:
            if self.dialog is not None and self.dialog.three_d_guide:
                xmin2, xmax2 = ax.get_xlim()
                tx  = xmin2 * 0.8 + xmax2 * 0.2
                ty  = ymin2 * 0.5 + ymax2 * 0.5
                ax.text(tx, ty, "See drift in 3D", alpha=0.2, fontsize=36)

        if peak_no is None:
            self.add_range_patches(ax, ymin2, ymax2, self.peak_eval_ranges)

        if draw_ranges:
            self.draw_ranges_impl(ax, mapper, peak_no)

    def draw_ranges_impl(self, ax, mapper, peak_no):
        ymin, ymax = ax.get_ylim()
        for k, range_ in enumerate( self.shifted_target_ranges ):
            lower, _, upper = range_
            for x in [lower, upper]:
                ax.plot( [ x, x ], [ ymin, ymax ], ':', color='black', alpha=0.2 )

            if peak_no is not None and k == peak_no:
                f, _, t = self.peak_eval_ranges[k]
                p = mpl_patches.Rectangle(
                        (f, ymin),      # (x,y)
                        t - f,          # width
                        ymax - ymin,    # height
                        facecolor   = 'cyan',
                        alpha       = 0.2,
                    )
                ax.add_patch( p )

    def get_baseline_color( self, mapper ):
        if self.anim_mode:
            color = 'yellow' if mapper.in_adj else 'red'
        else:
            color = 'red'
        return color

    def add_adjust_polygon(self, ax, j0, base, adj_base):
        x_end_i  = len(base) - 1
        vertices = [ (j0, base[0]), (j0+x_end_i, base[x_end_i]), (j0+x_end_i, adj_base[x_end_i]), (j0, adj_base[0]) ]
        polygon = mpl_patches.Polygon( vertices, alpha=0.3, fc='pink' )
        ax.add_patch(polygon)

    def draw_mapped(self, ax3, clear=False, title=True, draw_ranges=DRAW_MAPPING_RANGES, peak_no=None):
        if clear:
            ax3.cla()

        mapper = self.mapper

        if title:
            ax3.set_title( '(c)    Mapped elutions (Xray scattering scale)' + self.anim_counter_text, fontsize=self.title_fonsize )

        self.the_curves3 = []

        if self.anim_mode:
            ax2 = self.axes[1]
            ax3.set_xlim( ax2.get_xlim() )
            ax3.set_ylim( ax2.get_ylim() )

        if mapper.x_curve_y_adjusted is not None:
            ax3.plot( self.x3, mapper.x_curve_y_adjusted, color='orange' )
            self.the_curves3.append(["adjusted xray elution", self.x3, mapper.x_curve_y_adjusted])

            # self.the_curves3.append([])

        locally_showing = self.mapping_show_mode == 'locally'

        if locally_showing:
            show_vector = mapper.mapped_vector
        else:
            show_vector = mapper.get_uniformly_scaled_vector()

        if show_vector is not None:
            ax3.plot( self.x3, show_vector, color='blue' )
            self.the_curves3.append(["mapped uv elution", self.x3, show_vector])

        self.diff_y = show_vector - mapper.x_curve_y_adjusted
        self.show_hide_difference( update_state=False )

        cd_colors = self.dialog.get_cd_colors()
        for i, info in enumerate(mapper.x_curve.peak_info):
            peak    = info[1]
            if locally_showing:
                if mapper.mapped_spline is None:
                    peak_y  = None
                else:
                    peak_y  = mapper.mapped_spline(peak)
            else:
                peak_y  = show_vector[ int(peak + 0.5) ]
            if peak_y is not None:
                ax3.plot( self.xr_j0 + peak, peak_y, 'o', color=cd_colors[i] )

        for boudary in mapper.x_curve.boundaries:
            if locally_showing:
                if mapper.mapped_spline is None:
                    valley_y    = None
                else:
                    valley_y    = mapper.mapped_spline(boudary)
            else:
                valley_y    = show_vector[boudary]
            if valley_y is not None:
                ax3.plot( self.xr_j0 + boudary, valley_y, 'o', color=VALLEY_BOUNDARY_COLOR )

        base_color = self.get_baseline_color( mapper )
        ax3.plot( self.x3, np.zeros( len(mapper.x_curve.y) ), color=base_color )

        ymin, ymax = ax3.get_ylim()
        ax3.set_ylim( ymin, ymax )

        if peak_no is None:
            sci_text = False if self.using_xray_conc else True
            self.add_range_patches( ax3, ymin, ymax, self.peak_eval_ranges, sci_text=sci_text )

        if self.in_range_adjustment:
            xmin, xmax = ax3.get_xlim()
            xoffset = ( xmax - xmin )*0.12
            yoffset = ( ymax - ymin )*0.2

            ann_texts = [ 'S(%d)', 'P(%d)', 'E(%d)' ]
            y_  = mapper.x_curve.y
            for j, row in enumerate( self.dialog.annotation_points ):
                for k, p in enumerate( row ):
                    ax_ = p
                    ay_ = y_[p]
                    xoffset_ = xoffset * ( k - 1 ) + 1e-5       # + 1e-5 : workaroud for bug #12820
                    yoffset_ = yoffset if k == 1 else 0
                    ax3.annotate( ann_texts[k] % (p), xy=(ax_, ay_),
                                    xytext=( ax_ + xoffset_, ay_ - yoffset_ ),
                                    ha='center', va='center',
                                    arrowprops=dict( headwidth=5, width=0.5, color='black', shrink=0.05),
                                    )

        if draw_ranges:
            self.draw_ranges_impl(ax3, mapper, peak_no)

    def add_range_patches( self, ax, ymin, ymax, ranges, color=None, sci_text=False ):
        if ranges is None:
            # this case may occur in animation
            return

        if len(ranges) != len(self.sci_list):
            # as in 20180605/Backsub3
            self.mapper.logger.warning("found bug: len(ranges) != len(self.sci_list) as %d != %d", len(ranges), len(self.sci_list))

        for i, row in enumerate(ranges):
            if color is None and i < len(self.sci_list):
                sci = self.sci_list[i]
                if sci >= SCI_BOUNDARIES[1]:
                    fc  = 'cyan'
                    tfc = 'blue'
                elif sci >= SCI_BOUNDARIES[0]:
                    fc  = 'yellow'
                    tfc = 'green'
                else:
                    fc  = 'pink'
                    tfc = 'red'
            else:
                fc  = color

            f = row[0]
            t = row[2]
            p = mpl_patches.Rectangle(
                    (f, ymin),      # (x,y)
                    t - f,          # width
                    ymax - ymin,    # height
                    facecolor   = fc,
                    alpha       = 0.2,
                )
            ax.add_patch( p )
            if sci_text:
                tx = (f + t)/2
                h = (ymax - ymin) * 0.03
                ax.text( tx, h, '%d' % int(sci), ha='center', fontsize=36, color=tfc, alpha=0.2 )

    def save_the_figure( self, folder, analysis_name ):
        # print( 'save_the_figure: ', folder, analysis_name )
        filename = analysis_name.replace( 'analysis', 'figure' )
        path = os.path.join( folder, filename )
        self.fig.savefig( path )

    def show_hide_difference( self, update_state=True ):

        if update_state:
            self.diff_visible = (self.diff_visible + 1) % 3

        self.plot_difference(draw_base=True)
        self.show()

        text = diff_btn_texts[ (self.diff_visible + 1) % 3]
        self.show_diff_btn.config( text=text )

    def update_show_diff_btn_state(self):
        state = Tk.NORMAL if self.mapping_show_mode == 'locally' else Tk.DISABLED
        self.show_diff_btn.config(state=state)

    def plot_difference(self, draw_base=False, debug=False):
        if self.ax3t is not None:
            self.ax3t.remove()
            self.ax3t = None

        if debug:
            self.logger.info("plot_difference: diff_visible=%d", self.diff_visible)

        if self.diff_visible == 0:
            return

        from matplotlib.ticker import FormatStrFormatter

        fc_xe = self.mapper.get_mapped_flow_changes()

        slice_ = slice(*fc_xe)

        self.ax3t = self.axes[-1].twinx()
        amaxy = np.max(self.diff_y) * 1.1
        self.ax3t.set_ylim( -amaxy, amaxy )
        self.ax3t.yaxis.set_major_formatter(FormatStrFormatter('%.1e'))

        x = self.xr_j0 + self.mapper.x_curve.x
        y = self.diff_y if self.diff_visible == 1 else abs(self.diff_y)
        x_ = x[slice_]
        y_ = y[slice_]
        self.ax3t.bar( x_, y_, color='purple', alpha=0.3 )
        self.ax3t.plot( x[[0, -1]], [0, 0], color='pink' )

        diff_rec = ["difference", x, y]
        if len(self.the_curves3) < 3:
            self.the_curves3.append(diff_rec)
        else:
            self.the_curves3[-1] = diff_rec

        if draw_base:
            base_color = self.get_baseline_color( self.mapper )
            self.axes[-1].plot( x, np.zeros( len(x) ), color=base_color )

    def on_mpl_button_press(self, event):
        if event.xdata is None:
            return

        if event.button == 3:
            self.create_popup_menu(event)
            dialog = self.dialog
            rootx = dialog.winfo_rootx()
            rooty = dialog.winfo_rooty()
            w, h, x, y = split_geometry(self.mpl_canvas_widget.winfo_geometry())
            self.popup_menu.post(rootx + int(event.x), rooty + h - int(event.y))
            return

    def create_popup_menu(self, event):
        if self.popup_menu is None:
            self.popup_menu = Tk.Menu( self.dialog, tearoff=0 )
            self.popup_menu.add_command( label='Conc. Dependency', command=self.show_cdi_dialog )
            self.popup_menu.add_command( label='Save the curves', command=self.save_the_curves )

        i = self.axes.index(event.inaxes)
        state = Tk.DISABLED if i == 0 else Tk.NORMAL
        self.popup_menu.entryconfigure(0, state=state)
        self.last_click_event = event

    def show_cdi_dialog(self, pno=None, show_later=False):
        from Conc.CdInspection import CdInspectionDailog

        if pno is None:
            event = self.last_click_event
            i = self.axes.index(event.inaxes)
            assert i > 0
            x = event.xdata - self.xr_j0
            min_dist = None
            pno = None
            range_ = None
            for k, rec in enumerate(self.mapper.x_curve.peak_info):
                dist = abs(rec[1] - x)
                if min_dist is None or dist < min_dist:
                    min_dist = dist
                    pno = k
                    range_ = (rec[0], rec[2])
        else:
            i = 1
            rec = self.mapper.x_curve.peak_info[pno]
            range_ = (rec[0], rec[2])

        sd = self.dialog.mapper.baseline_corrected_copy
        if sd is None:
            sd = self.dialog.serial_data
        else:
            print('using corrected sd')
        q = sd.qvector
        f, t = range_
        eslice = slice(f, t+1)
        M = sd.intensity_array[eslice,:,1].T
        E = sd.intensity_array[eslice,:,2].T
        c = self.mapper.x_curve.y[eslice]
        c_ = c/np.max(c)
        C = np.array([c_])
        xray_scale = sd.get_xray_scale()
        self.cdi_dialog = CdInspectionDailog(self.dialog, self.dialog, M, E, C, q, slice(f+self.xr_j0, t+self.xr_j0),
                                                xray_scale=xray_scale, plotter_info=[self, i, pno])
        if show_later:
            pass
        else:
            self.cdi_dialog.show()

    def save_the_curves(self):
        from CurveSaverDialog import CurveSaverDialog

        the_curves = None
        event = self.last_click_event
        k = int(event.x/FIGURE_WIDTH_POSITIONING)
        print("save_the_curves", k, event.x)
        the_curves = [self.the_curves1, self.the_curves2, self.the_curves3][k]

        curve_list = []
        for k, rec in enumerate(the_curves):
            label, x, y = rec
            curve = np.array([x, y]).T
            curve_list.append([label, curve, label + '.dat'])

        save_folder = '/'.join( [get_setting('analysis_folder'), 'mapping'] )

        dialog = CurveSaverDialog(self.dialog, curve_list, save_folder)
        w, h, x, y = split_geometry(self.dialog.geometry())
        x_ = x + 200 + FIGURE_WIDTH_POSITIONING * k
        dialog.geometry("+%d+%d" % (x_,  y+400))
        dialog.show()



"""
    StructureFactorBoundsDemo.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import sys
import os
import numpy as np
import seaborn
seaborn.set()

# execute this file with python this-path demo
if len(sys.argv) > 1 and sys.argv[1].find("demo") >= 0:
    this_dir = os.path.dirname(os.path.abspath( __file__ ))
    sys.path.append(this_dir + '/..')
    import molass_legacy.KekLib, SerialAnalyzer

import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Optimizer.StructureFactorBounds import StructureFactorBounds
from Theory.SolidSphere import phi

def demo(qv=None, lrf_info=None, gk_info=None):
    from importlib import reload
    import Optimizer.StructureFactorBounds
    reload(Optimizer.StructureFactorBounds)
    from .StructureFactorBounds import StructureFactorBounds

    if qv is None:
        qv = np.linspace(0.005, 0.4, 200)

    Rg = 35
    R = np.sqrt(5/3)*Rg

    pv = phi(qv, R)**2
    sv = 1 - phi(qv, 2*R)

    fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(12,8))
    ax11 = axes[0,0]
    ax12 = axes[0,1]
    ax21 = axes[1,0]
    ax22 = axes[1,1]

    ax11.set_title("Hard Sphere")

    ax11.set_yscale('log')
    ax11.plot(qv, pv, label="P(q)")
    ax11.plot(qv, pv*sv, label="P(q)*S(q)")
    ax11.legend()

    bound = 1/(qv*R)**2
    ax12.plot(qv, sv - 1, color="C1", label="S(q) - 1")
    ax12.plot(qv, bound, ":", color="red", label=r"$\frac{1}{(qR)^2}$")
    ax12.plot(qv, -bound, ":", color="red", label=r"$\frac{1}{(qR)^2}$")
    ax12.set_ylim(-1, 1)

    ax12.legend()

    if lrf_info is not None:
        assert gk_info is not None
        Pxr, Cxr, Puv, Cuv, mapped_UvD = lrf_info.matrices

        n = 1
        aq = Pxr[:,n]
        bq = Pxr[:,-1]
        cv = Cxr[n,:]
        k = np.argmax(cv)
        c = cv[k]
        min_aq = np.ones(len(aq)) * 1e-6
        sf1v = bq/np.max([min_aq, aq], axis=0)
        sf_bounds = StructureFactorBounds(qv, lrf_info, gk_info)
        bounds = sf_bounds.get_bounds()
        penalty = sf_bounds.compute_penalty(Pxr)
        print("penalty=", penalty)

        ax21.set_yscale('log')
        ax21.plot(qv, aq*c + bq*c**2)
        ax21.plot(qv, aq*c)
        ax22.plot(qv, sf1v, color="C1")
        ax22.plot(qv, bounds[0], ":", color="red")
        ax22.plot(qv, bounds[1], ":", color="red")
        ax22.set_ylim(-10, 10)

    fig.tight_layout()
    plt.show()

def demo_for_debugger(dialog, canvas):
    from Kratky.GuinierKratkyInfo import GuinierKratkyInfo

    print("demo_for_debugger")

    sd = dialog.optinit_info.sd
    qv = sd.qvector
    params = canvas.get_current_params()
    optimizer = canvas.fullopt
    lrf_info = optimizer.objective_func(params, return_lrf_info=True)
    gk_info = GuinierKratkyInfo(optimizer, params, lrf_info)

    with plt.Dp():
        demo(qv=qv, lrf_info=lrf_info, gk_info=gk_info)

def proof1():
    R = np.sqrt(5/3)*100
    Qmax = 0.5
    qR_max = 0.5*R
    qR_lim = 6
    x1 = np.linspace(0, qR_lim, 100)
    x2 = np.linspace(0, qR_max, 200)

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
        fig.suptitle("Proof of sin(x) - x cos(x) < 1.1x where x > 0", fontsize=20)

        ax1.set_title("Proof of sin(x) - x cos(x) < 1.1x where 0 < x < %.2g" % qR_lim, fontsize=16)
        ax2.set_title("Proof of sin(x) - x cos(x) < 1.1x where 0 < x < %.2g" % qR_max, fontsize=16)

        for ax, x in [(ax1, x1), (ax2, x2)]:
            y = np.sin(x) - x*np.cos(x)
            ax.plot(x, y, label=r"$y=sin(x) - x cos(x)$")
            ax.plot(x, 1.1*x, label=r"$y=1.1x$")
            ax.plot(x, -1.1*x, label=r"$y=-1.1x$")
            ax.legend()

        fig.tight_layout()
        plt.show()

if __name__ == '__main__':
    # demo()
    proof1()



"""
    OptimizerSettings.py

    Copyright (c) 2022-2025, SAXS Team, KEK-PF
"""
import logging
import os
from collections import OrderedDict
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from molass_legacy.Experiment.ColumnTypes import PORESIZE_ALLOWANCE
from .TheUtils import get_optimizer_folder

def get_settings_path(optimizer_folder=None):
    if optimizer_folder is None:
        optimizer_folder = get_optimizer_folder()
    return os.path.join(optimizer_folder, "opt_settings.txt")

OPT_DEFAULT_SETTINGS = None

# task: unify these values with those in SerialSettings
def delayed_settings_init():
    from molass_legacy.Experiment.DataUtils import get_columntype, get_default_num_plates
    global OPT_DEFAULT_SETTINGS, OPT_DEFAULT_DICT

    default_columntype = get_columntype()
    poresize = default_columntype.poresize

    OPT_DEFAULT_SETTINGS = [
        ("param_init_type", 0),
        ("bounds_type", 0),
        ("elution_model", 0),       # not get_setting("elution_model") for backward compatibility
        ("t0_upper_bound", None),
        ("uv_basemodel", 0),
        ("poreexponent", 0),        # set_setting("poreexponent", None) doesn't seem to work. why?
        ("ratio_interpretation", 0),
        ("separate_eoii", 0),
        ("separate_eoii_type", 0),
        ("separate_eoii_flags", []),
        ("apply_sf_bounds", 0),
        ("sf_bound_ratio", 1.0),
        ("apply_rg_discreteness", 0),
        ("rg_discreteness_unit", 1.0),
        ("apply_mw_integrity", 0),
        ("mw_integer_ratios", None),
        ("avoid_peak_fronting", 0),
        ("optimization_method", 0),

        # SEC parameters
        ("exclusion_limit", default_columntype.excl_limit),
        ("poresize", poresize),
        ("poresize_bounds", (poresize - PORESIZE_ALLOWANCE, poresize + PORESIZE_ALLOWANCE)),
        ("num_plates_pc", int(round(get_default_num_plates()*0.3)))
        ]

    OPT_DEFAULT_DICT = dict(OPT_DEFAULT_SETTINGS)

class OptimizerSettings:
    def __init__(self, **kwargs):
        if OPT_DEFAULT_SETTINGS is None:
            delayed_settings_init()

        self.logger = logging.getLogger(__name__)
        self._dict = OrderedDict()

        if len(kwargs) > 0:
            for key, value in OPT_DEFAULT_SETTINGS[0:2]:
                self._dict[key] =  kwargs.pop(key, value)
            assert len(kwargs) == 0
            for key, _ in OPT_DEFAULT_SETTINGS[2:]:
                self._dict[key] = get_setting(key)

    def save(self, path=None):
        if path is None:
            path = get_settings_path()
        with open(path, "w") as fh:
            fh.write(self.__repr__())

    def load(self, path=None, optimizer_folder=None):
        from molass_legacy.KekLib.EvalUtils import eval_file
        if path is None:
            path = get_settings_path(optimizer_folder=optimizer_folder)

        def replacer(code):
            return code.replace("OptimizerSettings", "OrderedDict")

        self._dict = eval_file(path,
                               locals_=globals(),   # for OrderedDict
                               replacer=replacer)

        for key, _ in OPT_DEFAULT_SETTINGS[2:]:
            value = self.get(key)
            # print("set_setting(%s, %s)" % (key, str(value)))
            set_setting(key, value)

        separate_eoii = self.get("separate_eoii")
        if separate_eoii:
            key = "separate_eoii_type"
            separate_eoii_type = self.get(key)
            if separate_eoii_type == 0:
                value = 1
                self._dict[key] = value
                set_setting(key, value)
                self.logger.info("%s has been set to %d over %d to make it backward compatible.", key, value, separate_eoii_type)

        self.logger.info("optimizer settings have been restored as %s", self.__repr__())

    def keys(self):
        return [k for k, v in OPT_DEFAULT_SETTINGS]

    def get(self, key):
        return self._dict.get(key, OPT_DEFAULT_DICT[key])  #

    def __repr__(self):
        return str(self._dict).replace("OrderedDict", "OptimizerSettings")

def get_advanced_settings_text():
    # task: make this complete
    ratio_interpretation = get_setting("ratio_interpretation")
    separate_eoii = get_setting("separate_eoii")
    separate_eoii_type = get_setting("separate_eoii_type")
    separate_eoii_flags = get_setting("separate_eoii_flags")
    apply_sf_bounds = get_setting("apply_sf_bounds")
    sf_bound_ratio = get_setting("sf_bound_ratio")
    avoid_peak_fronting = get_setting("avoid_peak_fronting")

    if separate_eoii_type == 2:
        flags_text = ": %s" % str(separate_eoii_flags)
    else:
        flags_text = ""

    if separate_eoii_type > 0:
        flags_text += ", bounded with %g" % sf_bound_ratio if apply_sf_bounds else ""

    ret_text = ""
    for flag, text in [ (ratio_interpretation, "rational interpretation"),
                        (separate_eoii, "separating effect of interparticle interactions" + flags_text),
                        (avoid_peak_fronting, "avoiding peak fronting"),
                        ]:
        if flag:
            if ret_text > "":
                ret_text += ", "
            ret_text += text

    return ret_text



"""
    PolygonGeometry.py

    borrowed from
    https://stackoverflow.com/questions/68104969/offset-a-parallel-line-to-a-given-line-python/68109283    
"""
import numpy as np

def normalizeVec(x,y):
    distance = np.sqrt(x*x+y*y)
    return x/distance, y/distance

def makeOffsetPoly(oldX, oldY, offset, outer_ccw = 1):
    """
    there seems to be a bug when outer_ccw = -1 for inner offset
    use a negetive offset instead
    """
    num_points = len(oldX)
    newX = []
    newY = []

    for curr in range(num_points):
        prev = (curr + num_points - 1) % num_points
        next = (curr + 1) % num_points

        vnX =  oldX[next] - oldX[curr]
        vnY =  oldY[next] - oldY[curr]
        vnnX, vnnY = normalizeVec(vnX,vnY)
        nnnX = vnnY
        nnnY = -vnnX

        vpX =  oldX[curr] - oldX[prev]
        vpY =  oldY[curr] - oldY[prev]
        vpnX, vpnY = normalizeVec(vpX,vpY)
        npnX = vpnY * outer_ccw
        npnY = -vpnX * outer_ccw

        bisX = (nnnX + npnX) * outer_ccw
        bisY = (nnnY + npnY) * outer_ccw

        bisnX, bisnY = normalizeVec(bisX,  bisY)
        bislen = offset /  np.sqrt((1 + nnnX*npnX + nnnY*npnY)/2)

        newX.append(oldX[curr] + bislen * bisnX)
        newY.append(oldY[curr] + bislen * bisnY)
    
    return newX, newY


# 
import os

def get_version(toml_only=False):
    """
    Retrieve the version of the package from pyproject.toml or importlib.metadata.

    This function prioritizes reading the version from pyproject.toml to ensure
    that the local repository version is used during development or testing.
    If pyproject.toml is not found, it falls back to using importlib.metadata
    to retrieve the version of the installed package.

    Parameters:
    ------------
    toml_only: bool, optional
        If True, the function strictly reads the version from pyproject.toml.
        This is crucial to avoid confusion about the version being used,
        which can lead to significant time loss during testing.
        (confusion about the local repository versus the installed).

        If False, the function attempts to read the version from pyproject.toml
        first. If pyproject.toml does not exist, which means you are using the
        installed package, it falls back to using importlib.metadata to retrieve
        this version.

    Raises:
    -------
    AssertionError:
        If toml_only is True but pyproject.toml is not found. This ensures
        that the function behaves predictably in cases where the local
        repository is not available.

    This docstring was improved in collaboration with GitHub Copilot.
    """

    pyproject_toml = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'pyproject.toml')
    if os.path.exists(pyproject_toml):
        # If you are using the local repository, read the version from pyproject.toml
        import toml
        pyproject_data = toml.load(pyproject_toml)
        return pyproject_data['project']['version']
    else:
        # Otherwise, use importlib.metadata to get it from the installed package
        assert toml_only is False, "toml_only is not expected in this context"
        import importlib.metadata
        return importlib.metadata.version(__package__)


"""
    Models/RateTheory/EDM.py

    Copyright (c) 2023-2025, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
from matplotlib.widgets import Slider, Button
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SecTheory.Edm import edm_func
from molass_legacy.QuickAnalysis.ModeledPeaks import recognize_peaks
from molass_legacy.Models.ElutionCurveModels import egha
from molass_legacy.Models.Tentative import Model
from molass_legacy.Models.ElutionModelUtils import compute_4moments, x_from_height_ratio_impl
from molass_legacy.KekLib.BasicUtils import Struct

VERY_SMALL_VALUE = 1e-8
MIN_CINJ = 1e-6
MAX_CINJ = 5.0

if False:
    save_reg_data_fh = open("reg-data.csv", "w")
else:
    save_reg_data_fh = None

def edm_impl(x, t0, u, a, b, e, Dz, cinj):
    return edm_func(x-t0, u, a, b, e, Dz, cinj)

def edm_full_impl(x, t0, u, a, b, e, Dz, cinj, cinit=0, c0=0.0001, tinj=2.0, L=30, z=30):
    return edm_func(x-t0, u, a, b, e, Dz, cinj, cinit, c0, tinj, L, z)

def debug_plot_with_sliders_impl(x, y, init_params, title=None):
    plot_params = init_params.copy()
    slider_specs = [    ("t0", -50, 200, plot_params[0]),
                        ("u", 0, 3, plot_params[1]),
                        ("a", 0, 2, plot_params[2]),
                        ("b", -4, 1, plot_params[3]),
                        ("e", 0, 2, plot_params[4]),
                        ("Dz", 0, 1, plot_params[5]),
                        ("cinj", 0, 3, plot_params[6]),
                        # ("tinj", 0, 10, init_params[7]),
                        ]

    with plt.Dp():
        fig, ax = plt.subplots()
        dp = plt.get_dp()
        if title is None:
            title = "debug_plot_with_sliders_impl"
        ax.set_title(title)
        ax.plot(x, y)

        edm_line, = ax.plot(x, edm_impl(x, *init_params))

        def slider_update(k, val):
            # print([k], "slider_update", val)
            plot_params[k] = val
            y_ = edm_impl(x, *plot_params)
            edm_line.set_data(x, y_)
            dp.draw()

        slider_axes = []
        sliders = []
        for k, (label, valmin, valmax, valinit) in enumerate(slider_specs, start=0):
            ax_ = fig.add_axes([0.15, 0.8 - 0.08*k, 0.25, 0.03])
            slider  = Slider(ax_, label=label, valmin=valmin, valmax=valmax, valinit=valinit)
            slider.on_changed(lambda val, k_=k: slider_update(k_, val))
            slider_axes.append(ax_)
            sliders.append(slider)

        def reset_slider_valules(event):
            for slider in sliders:
                slider.reset()
            dp.draw()

        axreset = fig.add_axes([0.9, 0.0, 0.1, 0.075])
        breset = Button(axreset, 'Reset')
        breset.on_clicked(reset_slider_valules)

        fig.tight_layout()
        applied = plt.show()
        if applied:
            params = np.array([slider.val for slider in sliders])
        else:
            params = init_params
    return params

def guess_params_from_sdm(x, y, n1, t1, Nd, t0, debug=True):
    # Nd ≡ N0
    M1 = t0 + n1*t1
    M2 = 2*n1*t1**2 + (M1)**2/Nd
    M3 = 6*n1*t1**2*(Nd*t1 + n1*t1 + t0)/Nd
    # Nd = Td**2/M2 
    Td = np.sqrt(Nd*M2)
    area = np.sum(y)
    print("M1, M2, Td, area=", M1, M2, Td, area)
    # u = 30/Td
    u = 0.5
    # task: compute a, b using M3
    a = 0.5
    b = -4.0
    e = 0.4
    Dz = 0.02
    cinj = 0.1
    """
    t0, u, a, b, e, Dz, cinj
    init_params= [ 5.e+01  5.e-01  5.e-01 -4.e+00  4.e-01  2.e-02  1.e-01]
    params= [ 5.00122462e+01  2.63754524e-01  5.29024183e-02 -3.80947097e+00   5.70061653e-01  2.04147889e-02  4.78377216e-01]
    """
    def objective(params):
        t0, u, a, b, e, Dz, cinj = params
        y_ = edm_impl(x, t0, u, a, b, e, Dz, cinj)
        return np.sum((y_ - y)**2)
    
    init_params = np.array([t0, u, a, b, e, Dz, cinj])
    if debug:
        print("init_params=", init_params)
        debug_plot_with_sliders_impl(x, y, init_params, title="guess_params_from_sdm: before minimize")
    res = minimize(objective, init_params)
    if debug:
        debug_plot_with_sliders_impl(x, y, res.x, title="guess_params_from_sdm: after minimize")
    return res.x

def guess_init_params_better(x, y, M):
    area = np.sum(y)
    score_list = []
    params_list = []
    for uscale in [3]:
        t0 = M[1]/2
        u = 30/M[1]*uscale
        a = 0.5
        if M[3] < 0:
            b = -4.0
        else:
            b = -4.0
        e = 0.4
        Dz = 0.02
        cinj = M[0]/2.0 * 0.2
        params = np.array([t0, u, a, b, e, Dz, cinj])
        params_list.append(params)
        y_ = edm_impl(x, *params)
        area_ = np.sum(y_)
        score = abs(1 - area_/area)

        if np.isnan(score):
            score = np.inf

        score_list.append(score)

    k = np.argmin(score_list)
    return params_list[k]

def guess(x, y, init_params=None, debug=False, debug_info=None):
    if debug:
        from importlib import reload
        import Models.RateTheory.RobustEDM
        reload(Models.RateTheory.RobustEDM)
    from molass_legacy.Models.RateTheory.RobustEDM import guess_init_params

    if debug:
        def debug_plot_with_sliders():
            return debug_plot_with_sliders_impl(x, y, init_params, title="guess: before minimize with sliders")


        def debug_plot_params(x, y, params, title):
            print("params=", params)
            wx, wy, wy_resid, range_ = debug_info
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2,figsize=(12,5))
                fig.suptitle(title, fontsize=20)
                ax1.set_title("Guessing Range", fontsize=16)
                ax2.set_title("Guess in the Range", fontsize=16)

                ax1.plot(wx, wy)
                ax1.plot(wx, wy_resid)
                ax1.plot(x, y)
                ymin, ymax = ax1.get_ylim()
                ax1.set_ylim(ymin, ymax)
                for px in x[[0,-1]]:
                    ax1.plot([px, px], [ymin, ymax], ":", color="gray")

                ax2.plot(x, y)
                ax2.plot(x, edm_impl(x, *params))
                fig.tight_layout()
                plt.show()

    if init_params is None:
        M = compute_4moments(x, y)
        # init_params = guess_init_params_better(x, y, M)
        init_params = guess_init_params(M)
        area = np.sum(y)
        y_i = edm_impl(x, *init_params)
        area_i = np.sum(y_i)
        ratio = area_i/area
        print("area ratio=", ratio)
        if debug:
            print("M=", M)
            print("init_params=", init_params)
            changed_params = debug_plot_with_sliders()
            print("changed_params=", changed_params)
            changed_y_i = edm_impl(x, *changed_params)
            changed_area_i = np.sum(changed_y_i)
            changed_ratio = changed_area_i/area
            print("changed area ratio=", changed_ratio)
            if ratio < 0.1:
                print("changed_params=", changed_params)
                init_params = changed_params
            debug_plot_params(x, y, init_params, "guess: before minimize")

    def objective(p):
        y_ = edm_impl(x, *p)
        return np.sum((y_ - y)**2)

    ret = minimize(objective, init_params)

    if debug:
        print("M=", M)
        debug_plot_params(x, y, ret.x, "guess: after minimize")

    return ret.x

def guess_multiple_impl(x, y, num_peaks, respect_egh=False, debug=False):
    egha_params_array = np.array(recognize_peaks(x, y, exact_num_peaks=num_peaks, affine=True))

    cy_list = []
    for params in egha_params_array:
        cy = egha(x, *params)
        cy_list.append(cy)

    if debug:
        ty = np.sum(cy_list, axis=0)
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("guess_multiple: EGH decomposition")
            ax.plot(x, y)
            for k, cy in enumerate(cy_list):
                ax.plot(x, cy, ":", label="component-%d" % k)
            ax.plot(x, ty, ":", color="red", lw=3, label="total")
            ax.legend()
            fig.tight_layout()
            plt.show()

    # sort in the order of peak heights
    exec_recs = sorted(np.array([np.arange(num_peaks), egha_params_array[:,0]]).T, key=lambda x: -x[1])
    print("exec_recs=", exec_recs)

    ecurve = Struct(x=x, y=y)   # dummy will sffice here
    edm_params_list = [None] * num_peaks
    y_resid = y.copy()
    edm_cy_list = [None] * num_peaks

    num_failures = 0
    for k, h_ in exec_recs:
        i = int(k)      
        egha_params = egha_params_array[i,:]
        try:
            ret_xes = x_from_height_ratio_impl(egha, ecurve, 0.1, *egha_params[1:])
            # print([i], egha_params, ret_xes)
        except:
            # as with pH6
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(None, "a possible component peak ignored due to x_from_height_ratio_impl failure: ")
            num_failures += 1
            continue

        range_ = np.logical_and(x > ret_xes[0], x < ret_xes[1])
        x_ = x[range_]
        if respect_egh:
            egh_cy = cy_list[i]
            y_ = egh_cy[range_]
        else:
            y_ = y_resid[range_]
        if debug:
            debug_info = (x, y, y_resid, range_)
        else:
            debug_info = None

        if debug:
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("guess_multiple: range")
                ax.plot(x, y_resid)
                ax.plot(x_, y_)
                fig.tight_layout()
                plt.show()

        sum_y = np.sum(y_)
        if sum_y < 0:
            # avoid computing for negative regions
            print("avoid computing for negative regions: np.sum(y_)=%g < 0" % sum_y)
            num_failures += 1
            continue

        params = guess(x_, y_, debug=debug, debug_info=debug_info)
        edm_params_list[i] = params
        if save_reg_data_fh is not None:
            j = np.argmax(y_)
            top_x = x_[j]
            top_y = y_[j]
            save_reg_data_fh.write(",".join(["%.3g" % v for v in [x_[0], top_x, top_y, *params]]) + "\n")
        cy = edm_impl(x, *params)
        edm_cy_list[i] = cy
        y_resid -= cy

        if debug:
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("guess_multiple: cy, y_resid")
                ax.plot(x_, y_)
                ax.plot(x, cy, ":")
                ax.plot(x, y_resid)
                fig.tight_layout()
                plt.show()

    if debug:
        print("num_failures=", num_failures)
        def draw_edm_cy_list(title, edm_cy_list):
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title(title)
                ax.plot(x, y)

                cy_list = []
                for k, cy in enumerate(edm_cy_list):
                    if cy is not None:
                        ax.plot(x, cy, ":", label="component-%d" % k)
                        cy_list.append(cy)

                ty = np.sum(cy_list, axis=0)
                ax.plot(x, ty, ":", color="red", lw=3, label="total")
                ax.legend()

                fig.tight_layout()
                ret = plt.show()
            return ret

        draw_edm_cy_list("guess_multiple: EDM decomposition before cinj optimization", edm_cy_list)
    # 
    def cinj_ovjective(p, return_cy_list=False):
        cy_list = []
        for i, params in enumerate(edm_params_list):
            params_ = params.copy()
            params_[6] = p[i]
            cy = edm_impl(x, *params_)
            cy_list.append(cy)
        if return_cy_list:
            return cy_list
        ty = np.sum(cy_list, axis=0)
        return np.sum((y - ty)**2)

    init_cinjs = [p[6] for p in edm_params_list]
    bounds = [(MIN_CINJ, MAX_CINJ)] * num_peaks
    ret = minimize(cinj_ovjective, init_cinjs, method="Nelder-Mead", bounds=bounds)
    edm_cy_list = cinj_ovjective(ret.x, return_cy_list=True)
    if debug:
        draw_edm_cy_list("guess_multiple: EDM decomposition after cinj optimization", edm_cy_list)

    peak_pos = []
    for i, params in enumerate(edm_params_list):
        params[6] = ret.x[i]
        m = np.argmax(edm_cy_list[i])
        peak_pos.append(x[m])
    sort_pairs = sorted(zip(peak_pos, edm_params_list), key=lambda x: x[0])
    final_params_list = [pair[1] for pair in sort_pairs]

    if num_failures > 0:
        assert not respect_egh
        print("try guess_multiple_impl respect_egh=True")
        try:
            return guess_multiple_impl(x, y, num_peaks, respect_egh=True)
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(None, "guess_multiple_impl respect_egh=True failed: ")

        ret_list = []
        for p in final_params_list:
            if p is not None:
                ret_list.append(p)
    else:
        ret_list = final_params_list

    return np.array(ret_list)

class EDM(Model):
    def __init__(self, **kwargs):
        super(EDM, self).__init__(edm_impl, **kwargs)

    def get_name(self):
        return "EDM"

    def is_traditional(self):
        return False

    def guess_multiple(self, x, y, num_peaks, debug=False):
        return guess_multiple_impl(x, y, num_peaks, debug=debug)

    def eval(self, params=None, x=None):
        return self.func(x, *params)

    def x_from_height_ratio(self, ecurve, ratio, params):
        return x_from_height_ratio_impl(edm_impl, ecurve, ratio, *params, needs_ymax=True, full_params=True)

    def get_params_string(self, params):
        return 't0=%g, u=%g, a=%g, b=%g, e=%g, Dz=%g, cinj=%g' % tuple(params)

    def adjust_to_xy(self, params_list, x, y, props=None, devel=False):
        if props is None:
            areas = []
            for p in params_list:
                cy = edm_impl(x, *p)
                areas.append(np.sum(cy))
            props = np.array(areas)/np.sum(areas)

        print("props=", props)

        if devel:
            def plot_with_params_list(title, params_list):
                with plt.Dp():
                    fig, ax = plt.subplots()
                    ax.set_title(title)
                    ax.plot(x, y)
                    cy_list = []
                    for p in params_list:
                        cy = edm_impl(x, *p)
                        cy_list.append(cy)
                        ax.plot(x, cy, ":")
                    ty = np.sum(cy_list, axis=0)
                    ax.plot(x, ty, ":", color="red")
                    fig.tight_layout
                    plt.show()
            plot_with_params_list("before conversion", params_list)

        params_array = np.array(params_list)

        def objective(p):
            cy_list = []
            areas = []
            for params in p.reshape(params_array.shape):
                cy = edm_impl(x, *params)
                cy_list.append(cy)
                areas.append(np.sum(cy))
            ty = np.sum(cy_list, axis=0)
            props_ = np.array(areas)/np.sum(areas)
            fv = np.sum((ty - y)**2) + np.sum((props_ - props)**2)
            return fv

        ret = minimize(objective, params_array.flatten(), method='Nelder-Mead')
        print("ret.success=", ret.success)
        converted_array = ret.x.reshape(params_array.shape)

        if devel:
            plot_with_params_list("after conversion", converted_array)
        return converted_array

def test_it_from_editor_frame(editor):
    frame = editor.get_current_frame()
    model = frame.model
    print("edm_text_from_editor_frame: ", model.get_name())
    params_list = []
    for rec in frame.opt_recs:
        params = rec.get_params()
        params_list.append(params)

    fx = frame.fx
    x = frame.x
    y = frame.y
    uv_y = frame.uv_y

    converted_list = model.adjust_to_xy(params_list, fx, uv_y)

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
        ax1.plot(x, uv_y, color="blue")
        for params in converted_list:
            cy = edm_impl(fx, *params)
            ax1.plot(x, cy, ":")
        ax2.plot(x, y, color="orange")
        for params in params_list:
            cy = edm_impl(fx, *params)
            ax2.plot(x, cy, ":")
        fig.tight_layout()
        plt.show()


"""
    OurGini.py

    from https://stackoverflow.com/questions/39512260/calculating-gini-coefficient-in-python-numpy
"""
import numpy as np

def gini(x):
    # (Warning: This is a concise implementation, but it is O(n**2)
    # in time and memory, where n = len(x).  *Don't* pass in huge
    # samples!)

    # Mean absolute difference
    mad = np.abs(np.subtract.outer(x, x)).mean()
    # Relative mean absolute difference
    rmad = mad/np.mean(x)
    # Gini coefficient
    g = 0.5 * rmad
    return g



# coding: utf-8
"""
    SAngler_WLS.py

    Copyright (c) 2016, Masatsuyo Takahashi, KEK-PF
"""
import numpy        as np

class SAngler_WLS_Result:
    def __init__( self, A, B, sigmaA2, sigmaB2 ):
        self.params = np.array( [ A, B ] )
        self.std_errors = np.sqrt( [ sigmaA2, sigmaB2 ] )
        self.sigmaA2    = sigmaA2
        self.sigmaB2    = sigmaB2

    def cov_params( self ):
        # this is diagonal only for proof purpose
        return np.diag( [ self.sigmaA2, self.sigmaB2 ] )

class SAngler_WLS:
    def __init__( self, y, x, w ):
        sum_w   = np.sum( w )
        wx_     = np.inner( w, x )
        x2_     = x**2
        wx2_    = np.inner( w, x2_ )
        delta   = sum_w*wx2_ - wx_**2
        wy_     = np.inner( w, y )
        wxy_    = np.inner( w, x*y )
        d_      = 1/delta
        self.A  = d_*( wx2_*wy_ - wx_*wxy_ )
        self.B  = d_*( sum_w*wxy_ - wx_*wy_ )

        N = y.shape[0]
        e = y - ( self.A + self.B * x )
        S = np.inner( w, e**2 )
        sigma2 = S / ( N - 2 )

        self.sigmaA2 = d_ * wx2_ * sigma2
        self.sigmaB2 = d_ * sum_w * sigma2

    def fit( self ):

        return SAngler_WLS_Result( self.A, self.B, self.sigmaA2, self.sigmaB2 )



"""
    MeasuredData.py

    Copyright (c) 2019-2021, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
from mpl_toolkits.mplot3d import Axes3D
from molass_legacy.DataStructure.XrayData import XrayData
from molass_legacy.DataStructure.UvData import UvData
from molass_legacy.Mapping.CurveSimilarity import CurveSimilarity

class MeasuredData:
    def __init__(self, in_folder, uv_folder=None, sd=None, pre_recog=None, debug=False):

        if sd is None:
            self.xr = XrayData(in_folder)

            if uv_folder is None:
                uv_folder = in_folder

            self.uv = UvData(uv_folder)
            self.in_folder = in_folder
        else:
            self.xr = XrayData(None, sd=sd)
            self.uv = UvData(None, sd=sd)

        if debug:
            from MatrixData import simple_plot_3d
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14,6), subplot_kw={'projection': '3d'})
            simple_plot_3d(ax1, self.uv.data)
            simple_plot_3d(ax2, self.xr.data)
            fig.tight_layout()
            plt.show()

        self.pre_recognize(pre_recog)

        if debug:
            self.plot()

    def plot(self, auto=False, save=False):
        plt.push()
        fig = plt.figure(figsize=(21,7))
        fig.suptitle(get_in_folder(), fontsize=16)
        ax1 = fig.add_subplot(131, projection='3d')
        ax2 = fig.add_subplot(132, projection='3d')
        ax3 = fig.add_subplot(133)
        self.uv.plot(ax=ax1)
        self.xr.plot(ax=ax2)
        self.plot_pre_sync(ax3)
        fig.tight_layout()
        fig.subplots_adjust(top=0.85)
        block = not auto
        pause = 2 if auto else None
        plt.show(block=block, pause=pause)
        if save:
            import os
            from molass_legacy._MOLASS.SerialSettings import get_setting
            path = os.path.join(get_setting('analysis_folder'), get_setting('analysis_name'))
            fig.savefig(path)
        plt.pop()

    def pre_recognize(self, pre_recog):
        if pre_recog is None:
            self.cs = CurveSimilarity(self.uv.e_curve, self.xr.e_curve)
        else:
            self.cs = pre_recog.cs
        self.set_restriction(pre_recog)

    def plot_pre_sync(self, ax):

        ax.set_title("Preliminary Mapping", fontsize=16)
        x = self.xr.e_curve.x
        xr_y = self.xr.e_curve.y
        ax.plot(x, xr_y, color='orange', label='Elution Curve in Xray Scattering')

        x_ = self.cs.get_extended_x()
        uv_y = self.cs.get_uniformly_mapped_a_curve(x=x_)
        ax.plot(x_, uv_y, ':', color='blue', label='Elution Curve in UV Absorbance')
        for fx in [self.xr.j_slice.start, self.xr.j_slice.stop]:
            # print('fx=', fx, x[-1])
            if fx is not None:
                ylim = ax.get_ylim()
                ax.set_ylim()
                ax.plot([fx, fx], ylim, ':', color='yellow', linewidth=3, label='Flow Change Boundary')

        ax.legend()

    def set_restriction(self, pre_recog):
        mapping_params = (self.cs.slope, self.cs.intercept)
        if pre_recog is None:
            flow_changes = self.uv.set_restriction(self.xr.e_curve)
        else:
            flow_changes = pre_recog.get_real_flow_changes()    # should be real_flow_changes to avoid misjudgement after mapping
            # print('flow_changes=', flow_changes)
            self.uv.set_restriction(self.xr.e_curve, flow_changes)
        self.xr.set_restriction(mapping_params, flow_changes)

    def get_real_flow_changes(self):
        j_slice = self.xr.j_slice
        return j_slice.start, j_slice.stop

    def get_xdata_for_mct(self):
        return self.xr.copy_for_mct()

    def update_picking_params(self, logger):
        from molass_legacy._MOLASS.SerialSettings import get_xray_picking, get_setting, set_setting
        nh = get_setting('num_points_intensity')//2     # nh = 5 in default
        n = self.xr.i_slice.start + nh
        qvector = self.xr.vector
        q_cur = get_xray_picking()
        q_new = qvector[n]
        if q_new > q_cur:
            q_new_r = round(q_new, 2)   # e.g., 0.0286 -> 0.03
            set_setting('x_ecurve_picking_q', q_new_r)
            logger.info('ecurve-picking Q has been updated from %.3g to %.3g', q_cur, q_new_r)
            ret = True
        else:
            ret = False
        return ret

    def get_sd(self, jslice=None, xr_jslice=None, uv_jslice=None):
        from SerialDataUtils import get_xray_files, get_uv_filename
        from SerialData import SerialData
        datafiles = get_xray_files(self.in_folder)

        xr = self.xr
        qvector = xr.vector
        data = xr.data
        error = xr.error.data
        if xr_jslice is None:
            xr_jslice = jslice
        if xr_jslice is None:
            start = 0
            stop = data.shape[1]
        else:
            start = xr_jslice.start
            if start is None:
                start = 0
            stop = xr_jslice.stop
            if stop is None:
                stop = data.shape[1]
        xray_array = np.array([np.vstack([qvector, data[:,j], error[:,j]]).T for j in range(start, stop)])
        datafiles_ = datafiles[start:stop]

        uv_file = get_uv_filename(self.in_folder)
        uv = self.uv

        if uv_jslice is None:
            uv_jslice = jslice
        if uv_jslice is None:
            start = 0
            stop = uv.data.shape[1]
        else:
            start = uv_jslice.start
            stop = uv_jslice.stop
            if start is None:
                start = 0
            if stop is None:
                stop = uv.data.shape[1]

        uv_array = uv.data[:,start:stop]
        print('get_sd: uv_array.shape=', uv_array.shape)
        lvector = uv.vector
        col_header = uv.col_header[start:stop]
        mtd_elution = None
        data_info   = [ datafiles_, xray_array, uv_array, lvector, col_header, mtd_elution ]
        sd = SerialData( self.in_folder, self.in_folder, conc_file=uv_file, data_info=data_info )
        return sd

    def create_mapper(self, **kwargs):
        from molass_legacy.Mapping.ElutionMapper import ElutionMapper
        from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
        sd = self.get_sd(**kwargs)
        mapper = ElutionMapper(sd)
        absorbance = mapper.absorbance
        pre_recog = PreliminaryRecognition(sd)
        absorbance.compute_base_curve(pre_recog, 1)     # baseline type should be one of 0, 1, 4
        mapper.optimize()
        return mapper



# coding: utf-8
"""
    FolderInfo.py

    Copyright (c) 2019-2020, SAXS Team, KEK-PF
"""
import os
import re

filename_re = re.compile(r'^(\w+_)(\d+)([^\d]+)')

class FolderInfo:
    def __init__(self, files, conc_file, comments):
        self.conc_file = conc_file
        self.num_files = len(files)
        self.comments = comments

        self.folder, sample_name = os.path.split(files[0])

        m = filename_re.search(sample_name)
        if m:
            # pattern will be like SAMPLE_#####_sub.dat
            self.pattern = m.group(1) + '#'*len(m.group(2)) + m.group(3)
            start = int(m.group(2))
            stop = start + self.num_files
        else:
            self.pattern = 'Irregular Name:' + sample_name
            start = 0
            stop = self.num_files

        self.num_slice = slice(start, stop)

    def get_files(self):
        files = []
        for k in range(self.num_slice.start, self.num_slice.stop):
            name = self.pattern.replace('#####', '%05d' % k)
            files.append(os.path.join(self.folder, name))
        return files

    def get_comments(self):
        return self.comments



"""
    Solvers.MCMC.TrialBridge.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Optimizer.FvScoreConverter import convert_score

def mcmc_trial_impl(caller):
    from importlib import reload
    import Optimizer.FuncReloadUtils
    reload(Optimizer.FuncReloadUtils)
    from molass_legacy.Optimizer.FuncReloadUtils import reload_optimizer
    print("mcmc_trial_impl")
    canvas = caller.canvas
    optimizer = canvas.fullopt
    init_params = canvas.get_current_params()
    temp_optimizer = reload_optimizer(optimizer, init_params=init_params)
    separete_params = temp_optimizer.split_params_simple(init_params)
    print("separete_params: ", separete_params)
    ret = temp_optimizer.solve(init_params, method="mcmc", niter=10)
    ret_params = temp_optimizer.to_real_params(ret.x)
    fv = temp_optimizer.objective_func(ret_params)
    sv = convert_score(fv)
    with plt.Dp(button_spec=["OK", "Cancel"]):
        fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18, 5))
        fig.suptitle("mcmc_trial_impl", fontsize=20)
        axt = ax2.twinx()
        axt.grid(False)
        ax1.set_title("UV Decomposition", fontsize=16)
        ax2.set_title("Xray Decomposition", fontsize=16)
        ax3.set_title("Objective Function Scores in SV=%.3g" % sv, fontsize=16)
        temp_optimizer.objective_func(ret_params, plot=True, axis_info=(fig, (ax1, ax2, ax3, axt)))
        fig.tight_layout()
        plt.show()


"""

    ScatteringBaseline.py

    scattering baseline solver

    Copyright (c) 2017-2025, SAXS Team, KEK-PF

"""
import numpy as np
from scipy import stats
from molass_legacy._MOLASS.SerialSettings import get_setting

PERCENTILE_FIRST        = 25
PERCENTILE_SECOND       = 25
PERCENTILE_FINAL        = 10
VERY_SMALL_SLOPE_RATIO  = 0.01
CONVERGENCE_RATIO       = 0.1
ALTERNATING_LIMIT_RATIO = 0.5

class ScatteringBaseline:
    def __init__( self, y, x=None, height=None, curve=None, logger=None, suppress_warning=False):
        if x is None:
            if logger is None:
                import logging
                logger = logging.getLogger(__name__)

            if curve is None:
                if not suppress_warning:
                    from molass_legacy.KekLib.DebugUtils import show_call_stack
                    show_call_stack("ScatteringBaseline")
                    logger.warning("*************** using ScatteringBaseline without x is deprecated.")
                x = np.arange(len(y))
            else:
                x = curve.x
                # logger.info("x has been set from the given curve")
        self.y = y
        self.x = x

        if height is None:
            height  = np.max( y ) - np.min( y )

        self.very_small_sploe = height / len(y) * VERY_SMALL_SLOPE_RATIO
        self.curve = curve
        self.logger = logger

    def solve( self, p_final=PERCENTILE_FINAL, max_iter_num=10, no_shift=False, debug=False ):
        xpp, ypp, slope, intercept, ppp, adopted_sice = self.get_low_percentile_params( self.x, self.y, max_iter_num, debug=debug )
        y_  = self.y - slope * self.x
        if adopted_sice is not None:
            y_  = y_[adopted_sice]
        p   = np.percentile( y_, p_final )
        p_  = np.where( y_ <= p )[0]
        m   = np.argmax( y_[p_] )
        n   = p_[m]

        A   = slope
        if no_shift:
            B   = intercept
        else:
            B   = self.y[n] - slope*self.x[n]
        self.xpp    = xpp
        self.ypp    = ypp
        self.npp    = n
        self.params = ( A, B )

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            x = self.x
            y = self.y
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("ScatteringBaseline.solve")
                ax.plot(x, y)
                ax.plot(x, y_)
                ax.plot(x, A*x + B, color="red")
                fig.tight_layout()
                plt.show()

        return A, B

    def is_alternating( self, slope_list ):
        slope_array = np.array(slope_list)
        half_iter_num = len(slope_list)
        average_slope = np.average(slope_array)
        slope_devs = slope_array - average_slope
        total_stdev = np.std(slope_array)
        upper_ratio = np.std(slope_array[slope_devs > 0])/total_stdev
        lower_ratio = np.std(slope_array[slope_devs < 0])/total_stdev
        # print('ratios=', upper_ratio, lower_ratio)
        alternating = (upper_ratio + lower_ratio) < ALTERNATING_LIMIT_RATIO
        # this is True for 20181203 (only as of Jan. 2019)
        return alternating, average_slope

    def debug_plot(self, x, y, y_, pc, title, ab=None):
        import molass_legacy.KekLib.DebugPlot as plt
        pp = y_ <= pc
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title(title)
            ax.plot(x, y)
            ax.plot(x, y_, ":")
            ax.plot(x[pp], y[pp], "o", color="yellow")
            if ab is not None:
                a, b = ab
                ax.plot(x, a*x + b, color="red")
            fig.tight_layout()
            plt.show()

    def get_low_percentile_params( self, x, y, max_iter_num, debug=False ):
        y_  = y
        ppp = np.percentile( y_, [ PERCENTILE_FIRST ] )
        last_slope  = None
        init_diff   = None
        alternating = False
        half_iter_num = max_iter_num//2
        slope_list = []

        if debug:
            self.debug_plot(x, y, y_, ppp[0], "get_low_percentile_params: entry")

        for i in range(max_iter_num):

            if i == half_iter_num:
                alternating, average_slope = self.is_alternating( slope_list )
                if alternating:
                    if self.logger is not None:
                        self.logger.warning( 'alternating state detected in LPM' )

            if alternating:
                if i > half_iter_num:
                    average_slope = np.average(slope_list[-half_iter_num:])
                slope = average_slope
            else:
                pp = y_ <= ppp[0]
                xpp = x[pp]
                ypp = y[pp]
                slope, intercept = stats.linregress( xpp, ypp )[0:2]

                if debug:
                    self.debug_plot(x, y, y_, ppp[0], "get_low_percentile_params: [%d] result" % i, ab=(slope, intercept))

            y_ = y - slope*x
            ppp = np.percentile( y_, [ PERCENTILE_SECOND ] )
            # print( [i], 'slope=', slope )

            slope_list.append(slope)

            if last_slope is not None:
                diff = abs( slope - last_slope )
                if init_diff is None:
                    init_diff = diff
                if ( diff  < self.very_small_sploe
                    or diff < init_diff * CONVERGENCE_RATIO
                    ):
                    break

            if alternating:
                # break
                pass

            last_slope = slope

        # using a list instead of a tuple for item asignment later
        ret = [xpp, ypp, slope, intercept, ppp, None]

        if self.curve is not None:
            peak_info = self.curve.peak_info
            rightmost_peak_foot_x = int(peak_info[-1][2])
            # print('peak_info=', peak_info, "rightmost_peak_foot_x=", rightmost_peak_foot_x)
            if xpp[0] > rightmost_peak_foot_x:
                if self.logger is not None:
                    self.logger.warning('retrying to replace the suspicious result in LPM')
                slice_ = slice(0, rightmost_peak_foot_x)
                x_ = y[slice_]
                y_ = y[slice_]
                retried_ret = self.get_low_percentile_params(x_, y_, max_iter_num, debug=debug)
                if abs(retried_ret[2]) < abs(ret[2]):
                    if self.logger is not None:
                        self.logger.warning('adopted a retried result in LPM')
                    retried_ret[-1] = slice_
                    return retried_ret

        if debug and self.logger is not None:
            self.logger.info('number of iterations was %d in LPM', i)

        return ret

    def demo_plot( self, title="Demo", parent=None, ecurve=None, p_final=None, hires=False ):
        import molass_legacy.KekLib.DebugPlot as plt
        A, B = self.params
        if p_final is None:
            p_final = PERCENTILE_FINAL

        figsize = (24,12) if hires else (10,8)
        fig = plt.figure(figsize=figsize)

        if hires:
            fontsize    = 48
            legendsize  = 36
            labelsize   = 36
            linewidth   = 5
            markersize  = 20
        else:
            fontsize    = 30
            legendsize  = 20
            labelsize   = 20
            linewidth   = 3
            markersize  = 8

        ax = fig.add_subplot( 111 )
        # ax.set_title( title, fontsize=fontsize )
        ax.plot( self.y, 'o', markersize=markersize )
        # ax.plot( self.y, ':', color='green' )
        ax.plot( self.xpp, self.ypp, 'o', markersize=markersize, color='orange', label='under %d%% points' % PERCENTILE_SECOND )

        baseline_ = A*self.x + B
        ax.plot( baseline_, color='red', linewidth=linewidth, alpha=1, label='LPM result with %d%% offset' % p_final )
        # ax.plot( self.npp, self.y[self.npp], 'o', color='red', label='10% point (default)' )

        ax.tick_params( labelsize=labelsize )
        ax.legend( fontsize=legendsize )
        fig.tight_layout()
        return plt.show()



"""

    SummaryBook.py

    Copyright (c) 2017-2024, SAXS Team, KEK-PF

"""
import os
import numpy                as np
import csv
import re
import logging
from datetime               import datetime
# from openpyxl               import Workbook
from openpyxl.styles        import Alignment
from openpyxl.styles.fonts  import Font
from openpyxl.utils         import get_column_letter
from openpyxl.chart         import BarChart, Reference, Series
from openpyxl.chart.data_source     import StrRef
from OpenPyXlUtil           import save_allowing_user_reply
from molass_legacy.AutorgKek.AppVersion import autoguinier_version_for_publication
from molass_legacy._MOLASS.Version import get_version_string, molass_version_for_publication
from molass_legacy.ATSAS.AtsasVersion import atsas_version_for_publication
from molass_legacy._MOLASS.SerialSettings import get_setting, SHOW_TEXT_DICT, get_beamline_name, get_xray_picking
from molass_legacy.UV.PlainCurveUtils import get_flat_wavelength
from LinearityScore         import linearity_score100, stderror_score100, FACTOR_WEIGHT
from TimeUtils              import seconds_to_datetime
from molass_legacy.Reports.DefaultFont import set_default_font
set_default_font()

NUM_SCORE_FACTORS   = 5
ZX_RATIO_MAX        = 0.97

def round_3g_str( v ):
    return 'NA' if v is None else '%.3g' % v

def round_3g( values ):
    return [ None if x is None else float( '%.3g' % x ) for x in values ]

def round_3g_with_error( values, errors ):
    return [ None if x is None else '%.3g±%.3g' % ( x, e ) for x, e in zip( values, errors ) ]

def np_product_safe( x, vec ):
    assert len( x.shape ) == 1
    assert len( x ) == len( vec )
    return np.array( [ None if v is None else v*vec[i] for i, v in enumerate( x ) ] )

class SummaryBook:
    def __init__( self, wb, controller ):
        self.logger = logging.getLogger( __name__ )
        self.parent = controller
        self.doing_sec = controller.doing_sec
        self.atsas_is_available = controller.atsas_is_available
        self.datgnom = controller.datgnom

        self.j0 = j0 = controller.xr_j0
        self.applied_ranges = controller.applied_ranges
        self.report_ranges = controller.report_ranges
        self.wb = wb
        self.ws_list = []
        self.has_extrapolation_result = len( controller.zx_summary_list ) > 0

        # As of openpyxl 2.4.8,
        # using this wb.active sheet causes a problem with EXCEL 2016.
        # wb.remove_sheet(wb.active)

        if self.has_extrapolation_result:
            ws = wb.create_sheet( "Extrapolation Summary" )
            self.ws_list.append( ws )
            self.make_zx_summary( ws, controller )

        ws = wb.create_sheet( "Entire Summary" )
        self.ws_list.append( ws )
        self.make_entire_summary( ws, controller )

        if self.has_extrapolation_result:
            ws = wb.create_sheet( "Summary for Publication" )
            self.ws_list.append( ws )
            self.make_publication_summary( ws, controller )

    def append(self, ws, row_list, alignments={}, number_formats={}, fonts={}, character_fonts={}, for_all_components=False, row_height=None ):
        self.row += 1
        ws.append( row_list )

        row = self.row

        for k, v in alignments.items():
            cell = ws.cell(row=row, column=k)
            if type(v) == int:
                cell.alignment = Alignment( indent=v )
            elif type(v) == str:
                cell.alignment = Alignment( horizontal=v )
            elif type(v) == dict:
                cell.alignment = Alignment( **v )
            else:
                assert False

        for k, v in number_formats.items():
            cell = ws.cell(row=row, column=k)
            cell.number_format = v

        for k, v in fonts.items():
            cell = ws.cell(row=row, column=k)
            cell.font = v

        if for_all_components and self.num_components > 1:
            ws.merge_cells(start_row=row, start_column=2, end_row=row, end_column=self.num_components+1)

        if row_height is not None:
            ws.row_dimensions[row].height = row_height

        if False:
            # openpyxl does not yet support these characters attr
            for k, v in character_fonts.items():
                cell = ws.cell(row=row, column=k)
                cell.characters[v[0]].font = v[1]

    def make_zx_summary( self, ws, controller  ):
        fontname = get_setting("report_default_font")

        # column lengths
        ws.column_dimensions['A'].width = 20
        ws.column_dimensions['B'].width = 24

        array1 = np.array( controller.zx_summary_list )
        zx_summary_array = np.hstack( [ array1, np.zeros( ( array1.shape[0], 1 + NUM_SCORE_FACTORS ) ) ] )

        zx_summary_array2 = np.array( controller.zx_summary_list2 )

        min_c = zx_summary_array[:,5]
        max_c = zx_summary_array[:,6]
        zx_ratio = ( max_c - min_c ) / max_c
        zx_summary_array[:,15] = zx_ratio

        zx_summary_array[:,16] = linearity_score100( zx_summary_array[:,11] )
        zx_summary_array[:,17] = stderror_score100(  zx_summary_array[:,12] )
        zx_summary_array[:,18] = linearity_score100( zx_summary_array[:,13] )
        zx_summary_array[:,19] = stderror_score100(  zx_summary_array[:,14] )
        zx_summary_array[:,20] = round_3g( np.min( [ np.ones( zx_ratio.shape ), zx_ratio/ZX_RATIO_MAX ], axis=0 )*FACTOR_WEIGHT )

        n_peak = int( np.max( zx_summary_array[:,0] ) )
        self.zx_summary_array = zx_summary_array
        self.zx_summary_array2 = zx_summary_array2
        self.n_peak = n_peak
        # print( 'n_peak=', n_peak )

        self.row    = 0
        peak_side_titles    = []
        peak_side_cols      = []
        peak_side_cells     = []
        peak_side_formulae  = []
        score_row   = 35
        range_type = get_setting('range_type')

        for i in range( zx_summary_array.shape[0] ):
            col = get_column_letter( i + 3 )
            ws.column_dimensions[col].width = 12
            m = int( zx_summary_array[i, 0] )
            j = zx_summary_array[i, 1]
            peak_num_ranges = controller.peak_num_ranges_list[m]

            if n_peak == 0:
                paren = ''
            else:
                paren = '(%d)' % (m+1)

            if self.doing_sec:
                if peak_num_ranges == 1 and range_type < 5:
                    comp = 'Both-sides'
                else:
                    comp = 'Asc-side' if j == 0 else 'Desc-side'
            else:
                comp = 'Nat-comp' if m == 0 else 'Unf-comp'
            comp_name = '%s%s' % ( comp, paren )
            peak_side_titles.append( comp_name )
            peak_side_cols.append( col )
            peak_side_cells.append( col+'1' )
            # TODO: compute 15, 18
            peak_side_formulae.append( '=SUM(%s%d:%s%d)' % ( col, score_row, col, score_row+4 ) )

        self.append( ws, [ None, None ] + peak_side_titles )
        for cell in peak_side_cells:
            ws[cell].alignment = Alignment(horizontal="center")

        factors = [ 'Rg Extrapolation Linearity', 'Rg Extrapolation StdError', 'I(0)/C Extrapolation Linearity',  'I(0)/C Extrapolation StdError', 'Extrapolation Ratio' ]

        self.append( ws, [ 'Extrapolation Range', 'From'              ] + list( self.j0 + zx_summary_array[:,2] ) )
        self.append( ws, [ None,                  'To'                ] + list( self.j0 + zx_summary_array[:,3] ) )
        self.append( ws, [ None,                  'Number of Points'  ] + list( zx_summary_array[:,4] ) )
        self.append( ws, [] )
        self.append( ws, [ 'Concentration',       'Minimum'           ] + round_3g( min_c ) )
        self.append( ws, [ None,                  'Maximum'           ] + round_3g( max_c ) )
        self.append( ws, [] )
        self.append( ws, [ 'Guinier Analysis',    'Rg'                ] + round_3g( zx_summary_array[:,7] ) )
        guinier_start_row = self.row
        self.append( ws, [ None,                  'Rg error'          ] + round_3g( zx_summary_array[:,8] ) )
        rg_results = controller.guinier_results
        basic_qualities = [result.basic_quality for result in rg_results]
        self.append( ws, [ None,                  'Rg basic quality'  ] + round_3g(basic_qualities) )
        self.append( ws, [ None,                  'Rg quality'        ] + round_3g([result.Quality for result in rg_results]) )
        self.append( ws, [ None,                  'qmin'              ] + round_3g([result.min_q for result in rg_results]) )
        self.append( ws, [ None,                  'qmax'              ] + round_3g([result.max_q for result in rg_results]) )
        self.append( ws, [ None,                  'qRg max'           ] + round_3g([result.max_qRg for result in rg_results]) )
        self.append( ws, [ None,                  'I(0)/C'            ] + round_3g( zx_summary_array[:,9] ) )
        self.append( ws, [ None,                  'I(0)/C error'      ] + round_3g( zx_summary_array[:,10] ) )

        iz_max_c        = np_product_safe( zx_summary_array[:, 9],  max_c )
        iz_err_max_c    = np_product_safe( zx_summary_array[:,10],  max_c )
        self.append( ws, [ None,                  'I(0) (max C)'        ] + round_3g( iz_max_c ) )
        self.append( ws, [ None,                  'I(0) error (max C)'  ] + round_3g( iz_err_max_c ) )
        guinier_end_row = self.row

        guinier_warning = False
        for k, quality in enumerate(basic_qualities):
            if quality < 0.5:
                guinier_warning = True
                for i in range(guinier_start_row, guinier_end_row+1):
                    cell = ws.cell(row=i, column=3+k)
                    cell.font = Font(name=fontname, color='FF0000')

        if guinier_warning:
            row = guinier_start_row + 2
            column = 4 + len(basic_qualities)
            for i, line in enumerate([
                "For Rg basic quality < 0.5, qmax may have been intentionally widened",
                "to get a possibly better Rg value resulting in an undesired qRg max > 1.3,",
                "which violates the recommended guideline.",
                "See Section 5.1.1 of the User's Guide for details.",
                ]):
                cell = ws.cell(row=row + i, column=column)
                cell.value = line
                cell.font = Font(name=fontname, color='FF0000')

            self.logger.warning("Summary includes low quality results. Be aware.")

        if self.atsas_is_available:
            pr_rgs = round_3g( zx_summary_array2[:,0] )
            pr_rg_errors = round_3g( zx_summary_array2[:,1] ) 
            pr_qmins = round_3g( zx_summary_array2[:,5] )
            pr_qmaxes = round_3g( zx_summary_array2[:,6] ) 
            pr_izs = round_3g( zx_summary_array2[:,2] )
            pr_iz_errors = round_3g( zx_summary_array2[:,3] )
            pr_dmaxes = round_3g( zx_summary_array2[:,4] )
        else:
            pr_rgs = []
            pr_rg_errors = []
            pr_qmins = []
            pr_qmaxes = []
            pr_izs = []
            pr_iz_errors = []
            pr_dmaxes = []
        self.append( ws, [ None ])
        self.append( ws, [ 'P(r) Analysis',       'Rg'       ] + pr_rgs )
        self.append( ws, [ None,                  'Rg error' ] + pr_rg_errors )
        self.append( ws, [ None,                  'qmin'       ] + pr_qmins )
        self.append( ws, [ None,                  'qmax'       ] + pr_qmaxes )
        self.append( ws, [ None,                  'I(0)/C'       ] + pr_izs )
        self.append( ws, [ None,                  'I(0)/C error' ] + pr_iz_errors )
        self.append( ws, [ None,                  'Dmax'                ] + pr_dmaxes )
        self.append( ws, [] )
        self.append( ws, [ 'Quality Measurement', factors[0]          ] + round_3g( zx_summary_array[:,11] ) )
        self.append( ws, [ None,                  factors[1]          ] + round_3g( zx_summary_array[:,12] ) )
        self.append( ws, [ None,                  factors[2]          ] + round_3g( zx_summary_array[:,13] ) )
        self.append( ws, [ None,                  factors[3]          ] + round_3g( zx_summary_array[:,14] ) )
        self.append( ws, [ None,                  factors[4]          ] + round_3g( zx_summary_array[:,15] ) )
        self.append( ws, [] )
        self.append( ws, [ 'Quality Score',       factors[0]          ] + list( zx_summary_array[:,16] ) )
        assert self.row == score_row
        self.append( ws, [ None,                  factors[1]          ] + list( zx_summary_array[:,17] ) )
        self.append( ws, [ None,                  factors[2]          ] + list( zx_summary_array[:,18] ) )
        self.append( ws, [ None,                  factors[3]          ] + list( zx_summary_array[:,19] ) )
        self.append( ws, [ None,                  factors[4]          ] + list( zx_summary_array[:,20] ) )
        self.append( ws, [ None,                  'Total Score'         ] + peak_side_formulae  )

        for row in range(score_row,score_row+5):
            for col in peak_side_cols:
                ws[ col+str(row) ].number_format = '0.0'

        # Stacked Bar Chart
        num_cols = len(peak_side_cols)

        c_ = BarChart()
        c_.title    = "Quality Scores"
        c_.type     = "col"
        c_.style    = 11
        c_.grouping = "stacked"
        c_.overlap  = 100
        categories  = Reference(ws, min_col=3, max_col=2+num_cols, min_row=1)
        for k, factor in enumerate(factors):
            values = Reference(ws, min_col=3, max_col=2+num_cols, min_row=score_row+k)
            series = Series(values, title=factor)
            c_.series.append(series)
        c_.set_categories(categories)
        c_.y_axis.scaling.max = 100

        chart_start_pos = get_column_letter( 4 + num_cols ) + '21'
        ws.add_chart(c_, chart_start_pos)

    def make_entire_summary( self, ws, controller ):

        opt_params = controller.mapped_info.opt_params
        if False:
            self.logger.info('opt_params=' + str(opt_params))

        # column lengths
        ws.column_dimensions['A'].width = 40
        ws.column_dimensions['B'].width = 40
        ws.column_dimensions['C'].width = 60

        self.row    = 0

        # Column Title
        self.append( ws, [ 'Category',                    'Item Name',                                'Item Value' ] )

        # Input Data
        self.append( ws, [ 'Input Data',                'Xray Scattering Data Folder',              get_setting( 'in_folder' ) ] )
        self.append( ws, [ None,                        'Xray Scattering Data Shape',               str(controller.serial_data.intensity_array.shape) ] )
        self.append( ws, [ None,                        'UV Absorbance Data Folder/File',           get_setting( 'uv_folder' ) + '/' + get_setting( 'uv_file' ) ] )
        self.append( ws, [ None,                        'UV Absorbance Data Shape',                 str(controller.serial_data.conc_array.shape) ] )
        self.append( ws, [] )

        # Output Data
        self.append( ws, [ 'Output Data',               'Analysis Result Folder/Book',              get_setting( 'analysis_folder' )+ '/' + get_setting( 'result_book' ) ] )
        self.append( ws, [ None,                        'Averaged Data Folder',                     get_setting( 'averaged_data_folder' ) ] )
        self.append( ws, [ None,                        'Extrapolated Data Folder',                 controller.zx_out_folder ] )
        self.append( ws, [] )
        self.append( ws, [ 'Input Data Pre-processing', 'Number of Elution Points for Averaging',   get_setting( 'num_curves_averaged' ) ], alignments={ 3:'left' } )
        self.append( ws, [] )

        # Concentration Factors
        self.append( ws, [ 'Concentration Factors',     'Path Length Factor',                       get_setting( 'path_length' ) ], alignments={ 3:'left' }, number_formats={3:'0.0000'} )
        self.append( ws, [ None,                        'Extinction Coefficient',                   get_setting( 'extinction' ) ],  alignments={ 3:'left' }, number_formats={3:'0.0000'} )
        self.append( ws, [] )

        # UV/Xray Mapping Settings
        self.append( ws, [ 'UV/Xr-ay Mapping Settings', 'UV Absorbance Picking Wavelength λ₁ (nm)', get_setting('absorbance_picking')], alignments={ 3:'left' }, number_formats={3:'0.0'} )
        self.append( ws, [ None,                        'UV Absorbance Baseline Wavelength λ₂ (nm)', get_flat_wavelength(controller.serial_data.lvector) ], alignments={ 3:'left' }, number_formats={3:'0.0'} )
        uv_correction = opt_params.get_uv_correction_str()
        self.append( ws, [ None,                        'UV Absorbance Baseline Correction', uv_correction] )
        self.append( ws, [ None,                        'Xray Picking Scattering Vector q (Å⁻¹)',     get_xray_picking()], alignments={ 3:'left' }, number_formats={3:'0.000'} )
        xray_correction = opt_params.get_xray_correction_str()
        self.append( ws, [ None,                        'Xray Scattering Baseline Correction', xray_correction ] )
        self.append( ws, [] )

        # Zero-Concentration Analysis Settings
        matrix_formulation = get_setting('matrix_formulation')
        use_elution_models = get_setting('use_elution_models')
        if matrix_formulation:
            with_without = "with" if use_elution_models else "without"
            formulation_text = "Matrix formulation %s elution models" % with_without
        else:
            formulation_text = "Non-matrix formulation (iteration for each Q)"

        self.append( ws, [ 'Zero-Concentration Analysis Settings',  'Optimizer Formulation',  formulation_text  ], alignments={ 3:'left' } )

        text_dict   = SHOW_TEXT_DICT['range_type']
        range_type  = get_setting( 'range_type' )
        show_text   = text_dict[range_type]
        self.append( ws, [ None,                        'Range Determination Policy',  show_text  ], alignments={ 3:'left' } )

        ranges_str = ', '.join( [ str( range_ ) for range_ in self.report_ranges ]  )
        self.append( ws, [ None,                        'Analysis Ranges',                  ranges_str  ] )
        if not matrix_formulation:
            self.append( ws, [ None,                    'Width of Regression for Q-axis',   get_setting( 'zx_num_q_points' )  ], alignments={ 3:'left' } )

            zx_boundary_method = get_setting( 'zx_boundary_method' )
            if zx_boundary_method == 'FIXED':
                show_text   = 'Fixed at %g' % ( get_setting( 'zx_boundary' ) )
            else:
                text_dict   = SHOW_TEXT_DICT['zx_boundary_method']
                show_text   = text_dict[zx_boundary_method]
            self.append( ws, [ None,                    'Regression Boundary Method',       show_text  ] )
        self.append( ws, [] )

        # Analysis Quality
        self.append( ws, [ 'Analysis Quality',   'UV/Xray Mapping Adequacy',         None  ] )
        self.append( ws, [ None,                        'Xray Scattering Baseline Fitness', None  ] )
        self.append( ws, [] )

        # Time Info
        self.append( ws, [ 'Time Info',                 'Date/Time Created',                datetime.now() ], alignments={ 3:'left' } )
        self.append( ws, [ None,                        'Basesurface Correction',           seconds_to_datetime(controller.seconds_correction)], number_formats={3:'mm:ss'}, alignments={3:'left'} )
        self.append( ws, [ None,                        'Guinier Analysis Execution',       seconds_to_datetime(controller.seconds_guinier)], number_formats={3:'mm:ss'}, alignments={3:'left'} )
        self.append( ws, [ None,                        'Extrapolation Execution',          seconds_to_datetime(controller.seconds_extrapolation)], number_formats={3:'mm:ss'}, alignments={3:'left'} )
        self.append( ws, [] )

        # Other Info
        self.append( ws, [ 'Other Info',                'Program Version',          get_version_string() ] )
        env_info = controller.env_info
        self.append( ws, [ None,                        'Excel Version',            env_info.excel_version ] )
        self.append( ws, [ None,                        'Logical Number of CPUs',   os.cpu_count() ], alignments={3:'left'} )
        if env_info.gpu_info is not None:
            self.append( ws, [ None,                    'CUDA Tools Version',       env_info.cuda_tools_ver_str ] )

    def make_publication_summary( self, ws, controller ):
        fontname = get_setting("report_default_font")
        bold_font = Font(name=fontname, bold=True)
        gray_font = Font(name=fontname, color="C0C0C0")
        red_font = Font(name=fontname, color='FF0000')

        try:
            xray_info = self.get_xray_scattering_info()
        except Exception as exc:
            print( exc )
            xray_info = {}

        # q range
        serial_data = controller.serial_data
        qvector     = serial_data.qvector
        q_range     = '%.3g - %.3g' % tuple( qvector[ [0,-1] ] )    # note that qvector has been already trimmed

        # column lengths
        ws.column_dimensions['A'].width = 60

        zx_summary_array = self.zx_summary_array
        peak_names      = []
        peak_names_align  = {}
        conc_ranges     = []
        num_frames_list = []

        # self.num_components = len(self.applied_ranges)
        self.num_components = zx_summary_array.shape[0]     # here, the ascending side and descending side are treated as separate components

        k = 0
        for p, range_ in enumerate( self.applied_ranges ):
            fromto_list = range_.get_fromto_list()
            for j, ft in enumerate(fromto_list):
                col = get_column_letter(k+2)
                ws.column_dimensions[col].width = 17
                if len(fromto_list) == 1:
                    side = "bth"
                else:
                    side = "asc" if j == 0 else "dsc"
                peak_names.append( 'Component-%d (%s)' % (p+1, side) )
                conc_ranges.append( '%.3g - %.3g' % tuple(zx_summary_array[k, [5,6]]) )
                num_frames = ft[1] - ft[0] + 1
                num_frames_list.append(num_frames)
                peak_names_align[k+3] = 'center'
                k += 1

        self.row    = 0
        self.append( ws, [ 'A. Sample details',                 ], fonts={ 1:bold_font } )
        peak_names_align[1] = 1     # add an indent
        self.append( ws, [   'Sample name(s)' ] + peak_names, alignments=peak_names_align )
        self.append( ws, [   'Organism',                        ], alignments={ 1:1 } )
        self.append( ws, [   'Source',                          ], alignments={ 1:1 } )
        self.append( ws, [   'Uniprot ID (residues in construct)',   ], alignments={ 1:1 } )
        self.append( ws, [   'Extinction coefficient (A280nm, 0.1%(w/v))', get_setting( 'extinction' )  ], alignments={ 1:1, 2:'left' }, number_formats={2:'0.0000'}, for_all_components=True )
        self.append( ws, [   'P.S.V, ν from chemical composition (cm³ g⁻¹)' ], alignments={ 1:1 } )
        self.append( ws, [   'Contrast, Δρ (1010 cm⁻²)',      ], alignments={ 1:1 } )
        self.append( ws, [   'Theoretical MW (Da) ',            ], alignments={ 1:1 } )
        self.append( ws, [   'SEC-SAXS column',                 ], alignments={ 1:1 } )
        self.append( ws, [   'Loading conc. (mg ml⁻¹)',         ], alignments={ 1:1 } )
        self.append( ws, [   'Injection volume (μL)',          ], alignments={ 1:1 } )
        self.append( ws, [   'Flow rate (ml min⁻¹)',            ], alignments={ 1:1 } )
        self.append( ws, [   'Conc. range (mg ml⁻¹)',           ] + conc_ranges, alignments={ 1:1 } )
        self.append( ws, [   'Solvent composition',             ], alignments={ 1:1 } )
        if get_setting('use_xray_conc'):
            conc_method = None
            for_all_components = None
        else:
            conc_method =  "UV-Visible spectroscopy"
            for_all_components = True
        self.append( ws, [   'Concentration method', conc_method ], alignments={ 1:1 }, for_all_components=for_all_components )
        self.append( ws, [] )

        self.append( ws, [ 'B. SAS data collection',            ], fonts={ 1:bold_font } )

        if False:
            self.append( ws, [  'Beam geometry (μm)',              xray_info.get( 'Beam geometry' )  ], alignments={ 1:1 } )
            self.append( ws, [  'Wavelength (Å)',                   xray_info.get( 'Wavelength' )  ], alignments={ 1:1, 2:'left' } )
            self.append( ws, [  'Exposure time (s)',                xray_info.get( 'Exposure time' )  ], alignments={ 1:1, 2:'left' } )
            self.append( ws, [  'Temperature (K)',                  ], alignments={ 1:1 } )

        self.append( ws, [   'Instrument', get_beamline_name(formal=True)  ], alignments={ 1:1 }, for_all_components=True )
        self.append( ws, [   'Wavelength (Å)',                  ], alignments={ 1:1 } )
        self.append( ws, [   'Beam geometry (μm)',             ], alignments={ 1:1 } )
        self.append( ws, [   'Camera length (mm)',              ], alignments={ 1:1 } )
        self.append( ws, [   'Q range (Å⁻¹)', q_range  ], alignments={ 1:1 }, character_fonts={ 1:(slice(0,1),bold_font) }, for_all_components=True )
        self.append( ws, [   'Absolute scaling method', 'Comparison with scattering from 1 mm pure H₂O'  ], alignments={ 1:1 }, for_all_components=True )
        self.append( ws, [   'Exposure time (sec), No. of frames' ] + [" , %d" % n for n in num_frames_list], alignments={ 1:1 } )
        self.append( ws, [   'Sample path length (mm)',         1, ], alignments={ 1:1, 2:'left' }, number_formats={2:'0.00'}, for_all_components=True )
        self.append( ws, [   'Sample temperature (K)',          ], alignments={ 1:1 } )
        if self.num_components > 1:
            alignments = { 1:{'indent':1, 'vertical':'center'}, 2:{'wrap_text':True} }
            row_height = 29.0
        else:
            alignments = { 1:1 }
            row_height = None
        self.append( ws, [   'Normalization', 'To incident and transmitted intensities measured with a micro ion chamber and a beam stopper with embedded photodiode'  ],
                                                                    alignments=alignments, for_all_components=True, row_height=row_height )
        self.append( ws, [   'Method for monitoring radiation damage',   'Data frame-by-frame comparison'  ], alignments={ 1:1 }, for_all_components=True )
        self.append( ws, [] )

        self.append( ws, [ 'C. Software employed for SAS data reduction, analysis and interpretation',  ], fonts={ 1:bold_font } )
        sangler_version = get_setting('sangler_version')
        sangler_version_str = "Unknown" if sangler_version is None else "SAngler " + sangler_version
        program_versions = "%s, %s" %(sangler_version_str, molass_version_for_publication())
        self.append( ws, [  'SAS data processing',              program_versions  ], alignments={ 1:1 }, for_all_components=True )
        self.append( ws, [  'Extinction coefficient estimate',          'ProtParam' ], alignments={ 1:1 }, fonts={ 2:gray_font } )
        self.append( ws, [  'Calculation of contrast and PSV values',   'MULCh'     ], alignments={ 1:1 }, fonts={ 2:gray_font } )
        if self.atsas_is_available:
            datgnom_path = self.datgnom.exe_path
            datgnom_re = re.compile(r"(datgnom\w*)")
            m = datgnom_re.search(datgnom_path)
            exename = m.group(1)    # datgnom or datgnom4
            datgnom_str = ', %s (%s)'% (exename, atsas_version_for_publication())
        else:
            datgnom_str = ''
        program_versions = "%s from %s%s" % (autoguinier_version_for_publication(), molass_version_for_publication(), datgnom_str)
        self.append( ws, [  'Basic analyses: (Guinier, P(r))',  program_versions  ],
                                                            alignments={ 1:1 }, character_fonts={ 1:(slice(26,27),bold_font) },
                                                            for_all_components=True )
        self.append( ws, [  'Shape modeling',                   ], alignments={ 1:1 } )
        self.append( ws, [  'Atomistic modeling',               ], alignments={ 1:1 } )
        self.append( ws, [  '3D graphic model representations', ], alignments={ 1:1 } )
        self.append( ws, [] )

        self.append( ws, [ 'D. Structural parameters',  ], fonts={ 1:bold_font } )
        self.append( ws, [ 'Guinier Analysis',  ], alignments={ 1:1 }, fonts={ 1:bold_font } )
        self.append( ws, [   'I(0) (cm⁻¹)',                     ] + round_3g_with_error(zx_summary_array[:,9], zx_summary_array[:,10]), alignments={ 1:2 } )
        guinier_start_row = self.row
        self.append( ws, [   'Rg (Å)',                          ] + round_3g_with_error(zx_summary_array[:,7], zx_summary_array[:,8]), alignments={ 1:2 } )

        rg_results = controller.guinier_results
        q_ranges = [ "%s - %s" % (round_3g_str(result.min_q), round_3g_str(result.max_q)) for result in rg_results]
        self.append( ws, [   'Q range (Å⁻¹)',                   ] +  q_ranges, alignments={ 1:2 } )
        qrg_ranges = [ "%s - %s" % (round_3g_str(result.min_qRg), round_3g_str(result.max_qRg)) for result in rg_results]
        self.append( ws, [   'Q×Rg range',                     ] + qrg_ranges, alignments={ 1:2 } )
        self.append( ws, [   'M from I(0) (ratio to predicted value)',    ], alignments={ 1:2 } )
        guinier_end_row = self.row

        max_qRg_warning  = False
        max_qRg_list = [result.max_qRg for result in rg_results]
        for j, max_qRg in enumerate(max_qRg_list):
            if max_qRg > 1.3:
                max_qRg_warning = True
                for i in range(guinier_start_row, guinier_end_row+1):
                    cell = ws.cell(row=i, column=2+j)
                    cell.font = red_font
        if max_qRg_warning:
            cell = ws.cell(row=guinier_start_row+3, column=3+len(max_qRg_list))
            cell.font = red_font
            cell.value = 'See "Extrapolation Summary" sheet for quality warnings.'

        zx_summary_array2 = self.zx_summary_array2 

        self.append( ws, [  'P(r) Analysis',                    ], alignments={ 1:1 }, fonts={ 1:bold_font })
        self.append( ws, [   'I(0) (cm⁻¹)',                     ] + round_3g_with_error(zx_summary_array2[:,2], zx_summary_array2[:,3]), alignments={ 1:2 } )
        self.append( ws, [   'Rg (Å)',                          ] + round_3g_with_error(zx_summary_array2[:,0], zx_summary_array2[:,1] ), alignments={ 1:2 } )
        alignments={ 1:2 }
        for k in range(self.num_components):
            alignments[k+2] = 'left'
        self.append( ws, [   'Dmax (Å)',                        ] + round_3g(zx_summary_array2[:, 4]), alignments=alignments )
        qranges = []
        if self.atsas_is_available:
            for qmin, qmax in zip(zx_summary_array2[:,5], zx_summary_array2[:,6]):
                qranges.append('%.4g - %.4g' % (qmin, qmax))
        self.append( ws, [   'Q range(Å⁻¹)',                    ] + qranges, alignments={ 1:2 } )
        self.append( ws, [   'M from I(0) (ratio to predicted value)',    ], alignments={ 1:2 } )
        self.append( ws, [   'Porod volume (Å³) (ratio to predicted)',  ], alignments={ 1:2 } )
        self.append( ws, [   'M from Bayesian Inference',       ], alignments={ 1:2 } )

        self.append( ws, [] )
        self.append( ws, [] )

        self.append( ws, [ '※This table was constructed with reference to the 2017 BioSAXS data publication guidelines.\n(https://doi.org/10.1107/S2059798317011597)', ],
                        alignments={ 1:1, 1:{'wrap_text':True} }, fonts={ 1:bold_font } )

    def get_xray_scattering_info( self ):
        ret_dict = {}
        in_folder = get_setting( 'in_folder' )
        info_file = in_folder + '/Xray_Scattering_Info.txt'
        fh = open( info_file )
        wavelength_re   = re.compile( r'Wavelength:\s+(\d+\.\d+)' )
        pixel_size_re   = re.compile( r'PixelSize:\s+(\d+e-\d+)\s*m\s+x\s+(\d+e-\d+)\s*m' )
        expos_time_re   = re.compile( r'Exposure Time:\s+(\d+\.\d+)\s*s'  )
        for line in fh:
            m =  wavelength_re.search( line )
            if m:
                wavelength = m.group(1)
                print( 'wavelength=', wavelength )
                ret_dict['Wavelength'] = float(wavelength)
                continue

            m =  pixel_size_re.search( line )
            if m:
                h_res = m.group(1)
                v_res = m.group(2)
                geometry = '%d × %d' % tuple([ int(float(x)*1e6) for x in [ h_res, v_res ] ])
                print( 'geometry=', geometry )
                ret_dict['Beam geometry'] = geometry
                continue

            m =  expos_time_re.search( line )
            if m:
                expos_time = m.group(1)
                print( 'expos_time=', expos_time )
                ret_dict['Exposure time'] = float(expos_time)
                continue

        fh.close()
        return ret_dict

    def add_format_setting(self, book_file):
        from .SummaryExcelFormatter import SummaryArgs

        ws = self.ws_list[-1]
        if ws.title.find("Publication") < 0:
            return

        args = SummaryArgs(ws.title, book_file)

        if self.parent.more_multicore:
            self.add_format_setting_more_multicore(args)
        else:
            self.add_format_setting_less_multicore(args)

    def add_format_setting_more_multicore(self, args):
        self.parent.teller.tell('summary_book', args=args)

    def add_format_setting_less_multicore(self, args):
        from .SummaryExcelFormatter import add_summary_format_setting
        add_summary_format_setting(self.parent.excel_client, args, self.logger)

    def save( self, xlsx_file ):
        save_allowing_user_reply( self.wb, xlsx_file )

    def save_as_csv( self, path ):
        """
            learned at
            https://stackoverflow.com/questions/10802417/how-to-save-an-excel-worksheet-as-csv-from-python-unix
        """
        for i, ws in enumerate( self.ws_list ):
            with open( path.replace( '.csv', '-%d.csv' % i ), 'w', encoding='utf-8', newline='') as f:
                c = csv.writer(f)
                for r in ws.rows:
                    row = [cell.value for cell in r]
                    c.writerow(row)



# coding: utf-8
"""

    ファイル名：   DeveloperOptions.py

    処理内容：

        開発者用の設定変更ダイアログ

    Copyright (c) 2017-2022, SAXS Team, KEK-PF

"""

import os
import platform
import wmi
import re
import warnings
import time

from molass_legacy.KekLib.OurTkinter             import Tk, Dialog, ttk, Font, FileDialog
from molass_legacy.KekLib.TkSupplements          import tk_set_icon_portable
import OurMessageBox        as MessageBox
from DevSettings            import get_dev_setting, set_dev_setting
from molass_legacy.KekLib.TkCustomWidgets        import FolderEntry
from molass_legacy.KekLib.BasicUtils             import get_home_folder
from molass_legacy._MOLASS.SerialSettings         import clear_settings, get_setting, set_setting
from MultiMonitor           import get_max_monitor, get_monitor_list, PYTHON_DEMO_MONITOR
from molass_legacy.KekLib.TkUtils                import adjusted_geometry

from molass_legacy.KekLib.ReadOnlyText           import ReadOnlyText
from MachineTypes           import is_on_virtual_machine, get_cpuid

ENABLE_HIRES_MODE_CHANGE    = False

class PlatformDialog(Dialog):
    def __init__(self, parent):
        Dialog.__init__(self, parent, "PlatformDialog" )

    def body( self, body_frame ):
        pinfo = ReadOnlyText(body_frame)
        pinfo.pack( fill=Tk.BOTH, expand=1, pady=10 )

        # Platform Info
        pinfo_list = [
            platform.machine(),
            platform.version(),
            str(platform.uname()),
            platform.system(),
            platform.processor(),
            ]

        for info in pinfo_list:
            pinfo.insert(Tk.END, info + '\n')

    def buttonbox( self ):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

class ProcessorDialog(Dialog):
    def __init__(self, parent):
        Dialog.__init__(self, parent, "ProcessorDialog" )

    def body( self, body_frame ):
        pinfo = ReadOnlyText(body_frame)
        pinfo.pack( fill=Tk.BOTH, expand=1, pady=10 )

        # Win32_Processor Info
        pinfo_list = [ str(s) for s in wmi.WMI().Win32_Processor()]

        for info in pinfo_list:
            pinfo.insert(Tk.END, info + '\n')

    def buttonbox( self ):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

class GpuDialog(Dialog):
    def __init__(self, parent):
        Dialog.__init__(self, parent, "GpuDialog" )

    def body( self, body_frame ):
        from Env.GpuInfo import DxdiagInfo

        ginfo = ReadOnlyText(body_frame)
        ginfo.pack( fill=Tk.BOTH, expand=1, pady=10 )

        if False:
            # DxdiagInfo
            # this takes some time
            dxinfo = DxdiagInfo()
            ginfo_list = dxinfo.get_info_list()

            for info in ginfo_list:
                ginfo.insert(Tk.END, info + '\n')

        # Win32_VideoController Info
        ginfo_list = [ str(s) for s in wmi.WMI().Win32_VideoController()]

        for info in ginfo_list:
            ginfo.insert(Tk.END, info + '\n')

    def buttonbox( self ):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)


class DeveloperOptionsDialog( Dialog ):
    def __init__( self, parent, title ):
        self.grab = 'local'     # used in grab_set
        self.parent             = parent
        self.title_             = title
        self.applied            = False

    def show( self ):
        self.parent.config( cursor='wait' )
        self.parent.update()

        Dialog.__init__(self, self.parent, self.title_ )

        self.parent.config( cursor='' )

    def body( self, body_frame_ ):   # overrides parent class method

        body_frame = Tk.Frame( body_frame_ )
        body_frame.pack( padx=40, pady=10 )

        tk_set_icon_portable( self )

        clear_button = Tk.Button( body_frame, text="Clear Setting", command=self.clear_settings )
        clear_button.pack( anchor=Tk.NW )

        # Caveate
        caveate = ( "Please note that those options below are temporary and not saved as permanent setting.\n"
                    + "In other words, you have to specify everytime after restart."
                  )
        msg = Tk.Label( body_frame, text=caveate, bg='white', fg='red' )
        msg.pack( fill=Tk.BOTH, expand=1, pady=10 )

        platform_frame = Tk.Frame(body_frame)
        platform_frame.pack()

        platform_btn = Tk.Button(platform_frame, text='Platform Info', command=lambda: PlatformDialog(self))
        platform_btn.grid(row=0, column=0, padx=10)

        processor_btn = Tk.Button(platform_frame, text='Processor Info', command=lambda: ProcessorDialog(self))
        processor_btn.grid(row=0, column=1, padx=10)

        cpuid_label = Tk.Label(platform_frame, text='ProcessorId: ' + str(get_cpuid()))
        cpuid_label.grid(row=0, column=2, padx=10)

        vitual_label = Tk.Label(platform_frame, text='Virtual Machine: ' + str(is_on_virtual_machine()))
        vitual_label.grid(row=0, column=3, padx=10)

        gpu_btn = Tk.Button(platform_frame, text='Gpu Info', command=lambda: GpuDialog(self))
        gpu_btn.grid(row=1, column=0, padx=10, pady=5, sticky=Tk.W)

        base_frame = Tk.Frame( body_frame );
        base_frame.pack( expand=1, fill=Tk.BOTH, padx=10, pady=10 )

        grid_row = -1

        # Display Selection
        grid_row += 1
        display_select_frame = Tk.Frame( base_frame )
        display_select_frame.grid( row=grid_row, column=0, sticky=Tk.W )

        display_select_label = Tk.Label( display_select_frame, text="Display Selection: " )
        display_select_label.grid( row=0, column=0, sticky=Tk.W )
        max_monitor = get_max_monitor()
        monitors = get_monitor_list()
        # print( 'monitors=', monitors )
        self.monitor = Tk.StringVar()
        env_monitor = os.environ.get( PYTHON_DEMO_MONITOR )
        if env_monitor is not None:
            self.monitor.set( env_monitor )
        for k, m in enumerate(monitors):
            m_ = str(m)
            if env_monitor is None and m is max_monitor:
                self.monitor.set( str(k) )
            rb = Tk.Radiobutton( display_select_frame, text=m_, variable=self.monitor, value=k )
            rb.grid( row=k+1, column=1, sticky=Tk.W )

        if ENABLE_HIRES_MODE_CHANGE:
            grid_row += 1
            self.hiresolution   = Tk.IntVar()
            cb = Tk.Checkbutton( base_frame, text="Hi-resolution mode for screenshots",
                                    variable=self.hiresolution  )
            cb.grid( row=grid_row, column=0, sticky=Tk.W )

        self.monitor_previous = self.monitor.get()

        # Taking Screenshots
        grid_row += 1
        self.take_screenshots = Tk.IntVar()
        self.take_screenshots.set( get_dev_setting( 'take_screenshots' ) )

        cb = Tk.Checkbutton( base_frame, text="Take screenshots automatically for window size problem investigation",
                                variable=self.take_screenshots  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W )

        grid_row += 1
        folder_frame = Tk.Frame( base_frame )
        folder_frame.grid( row=grid_row, column=0, sticky=Tk.W, padx=40 )

        folder_label = Tk.Label( folder_frame, text="Save Folder" )
        folder_label.grid( row=0, column=0, sticky=Tk.W )

        self.screenshot_folder = Tk.StringVar()
        folder = get_dev_setting( 'screenshot_folder' )
        if folder is None:
            folder = get_home_folder() + '/img'
        self.screenshot_folder.set( folder )
        folder_entry = FolderEntry( folder_frame, textvariable=self.screenshot_folder, width=60 )
        folder_entry.grid( row=0, column=1, sticky=Tk.W, padx=5 )

        grid_row += 1
        space = Tk.Label( base_frame, text="" )
        space.grid( row=grid_row, column=0 )

        # Intensity Data Reduction
        grid_row += 1
        self.intensity_reduction = Tk.IntVar()
        self.intensity_reduction.set( get_dev_setting( 'intensity_reduction' ) )

        cb = Tk.Checkbutton( base_frame, text="Reduce intensity data",
                                variable=self.intensity_reduction  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W )

        reduction_state = Tk.NORMAL

        grid_row += 1
        method_frame = Tk.Frame( base_frame )
        method_frame.grid( row=grid_row, column=0, sticky=Tk.W, padx=40 )

        method_grid_row = 0
        cycle_label = Tk.Label( method_frame, text="leaving one out of every " )
        cycle_label.grid( row=method_grid_row, column=0, sticky=Tk.E )

        self.reduction_cycle = Tk.IntVar()
        self.reduction_cycle.set( get_dev_setting( 'reduction_cycle' ) )
        self.reduction_cycle_entry = Tk.Spinbox( method_frame, textvariable=self.reduction_cycle,
                                            from_=2, to=4, increment=1, 
                                            justify=Tk.CENTER, width=6, state=reduction_state )
        self.reduction_cycle_entry.grid( row=method_grid_row, column=1, sticky=Tk.E )

        by_label = Tk.Label( method_frame, text=" Q-points by" )
        by_label.grid( row=method_grid_row, column=2, sticky=Tk.W )

        method_grid_row += 1
        self.reduction_method_buttons = []
        self.reduction_method = Tk.StringVar()
        self.reduction_method.set( get_dev_setting( 'reduction_method' ) )

        b = Tk.Radiobutton( method_frame, text='thinning out',
                    # font=self.fixed_font,
                    variable=self.reduction_method, value='THIN-OUT',
                    state=reduction_state,
                    )
        b.grid( row=method_grid_row, column=0, sticky=Tk.W, padx=20  )
        self.reduction_method_buttons.append( b )

        start_at_label = Tk.Label( method_frame, text="starting at " )
        start_at_label.grid( row=method_grid_row, column=1, sticky=Tk.W )

        self.reduction_start = Tk.IntVar()
        self.reduction_start.set( get_dev_setting( 'reduction_start' ) )
        self.reduction_start_entry = Tk.Spinbox( method_frame, textvariable=self.reduction_start,
                                            from_=0, to=3, increment=1, 
                                            justify=Tk.CENTER, width=6, state=reduction_state )
        self.reduction_start_entry.grid( row=method_grid_row, column=2, sticky=Tk.W )

        method_grid_row += 1
        b = Tk.Radiobutton( method_frame, text='averaging',
                    # font=self.fixed_font,
                    variable=self.reduction_method, value='AVERAGE',
                    state=Tk.DISABLED,
                    )
        b.grid( row=method_grid_row, column=0, sticky=Tk.W, padx=20 )
        self.reduction_method_buttons.append( b )

        # Show number of iterations entry for xray-scattering baseline corection
        grid_row += 1
        self.show_num_iterations = Tk.IntVar()
        self.show_num_iterations.set( get_dev_setting( 'show_num_iterations' ) )

        cb = Tk.Checkbutton( base_frame, text="Show number of iterations entry for xray-scattering baseline corection",
                                variable=self.show_num_iterations  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W )

        # Suppress usable_limit cutting
        grid_row += 1
        self.no_usable_q_limit = Tk.IntVar()
        self.no_usable_q_limit.set( get_dev_setting( 'no_usable_q_limit' ) )

        cb = Tk.Checkbutton( base_frame, text="Suppress usable_q_limit cutting",
                                variable=self.no_usable_q_limit  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W )

        # Log Memory Usage
        grid_row += 1
        self.log_memory_usage = Tk.IntVar()
        self.log_memory_usage.set( get_dev_setting( 'log_memory_usage' ) )

        cb = Tk.Checkbutton( base_frame, text="Log memory usage",
                                variable=self.log_memory_usage  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W )

        # Log Xray LPM Params
        grid_row += 1
        self.log_xray_lpm_params = Tk.IntVar()
        self.log_xray_lpm_params.set( get_dev_setting( 'log_xray_lpm_params' ) )

        cb = Tk.Checkbutton( base_frame, text="Log Xray LPM Params",
                                variable=self.log_xray_lpm_params  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W )

        # Show number of iterations entry for xray-scattering baseline corection
        grid_row += 1
        self.show_num_iterations = Tk.IntVar()
        self.show_num_iterations.set( get_dev_setting( 'show_num_iterations' ) )

        cb = Tk.Checkbutton( base_frame, text="Show number of iterations entry for xray-scattering baseline corection",
                                variable=self.show_num_iterations  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W )

        # Suppress boundary test deferring to allow boundaries in Guinier interval
        grid_row += 1
        self.suppress_defer_test = Tk.IntVar()
        self.suppress_defer_test.set( get_dev_setting( 'suppress_defer_test' ) )

        cb = Tk.Checkbutton( base_frame, text="Suppress boundary test deferring to allow boundaries to enter in the Guinier interval",
                                variable=self.suppress_defer_test  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W )

        # Use DATGNOM to fill [from P(r)] items in the Summary for Puiblication
        grid_row += 1
        self.use_datgnom = Tk.IntVar()
        self.use_datgnom.set( get_dev_setting( 'use_datgnom' ) )

        cb = Tk.Checkbutton( base_frame, text="Use DATGNOM to fill [from P(r)] items in the Summary for Puiblication",
                                variable=self.use_datgnom  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W )

        # Enable baseline drift simulation
        grid_row += 1
        self.enable_drift_simulation = Tk.IntVar()
        self.enable_drift_simulation.set( get_setting( 'enable_drift_simulation' ) )

        cb = Tk.Checkbutton( base_frame, text="Enable baseline drift simulation",
                                variable=self.enable_drift_simulation  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W )

        # Constant Term in Regression
        grid_row += 1
        self.zx_add_constant = Tk.IntVar()
        self.zx_add_constant.set( get_dev_setting( 'zx_add_constant' ) )

        cb = Tk.Checkbutton( base_frame, text="Add a constant term in Q-iterated WLS for A(q), B(a)",
                                variable=self.zx_add_constant  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W )

        # Constant Row in Concentration Matrix
        grid_row += 1
        self.add_conc_const = Tk.IntVar()
        self.add_conc_const.set( get_dev_setting( 'add_conc_const' ) )

        cb = Tk.Checkbutton( base_frame, text="Add a constant row to concentration matrix",
                                variable=self.add_conc_const )
        cb.grid( row=grid_row, column=0, sticky=Tk.W )

        # Individual B(q) Ignore
        grid_row += 1
        self.individual_bq_ingore = Tk.IntVar()
        self.individual_bq_ingore.set( get_dev_setting( 'individual_bq_ingore' ) )

        cb = Tk.Checkbutton( base_frame, text="Enable individual bq_ingore",
                                variable=self.individual_bq_ingore )
        cb.grid( row=grid_row, column=0, sticky=Tk.W )

        # Excel Visibility
        grid_row += 1
        self.make_excel_visible = Tk.IntVar()
        self.make_excel_visible.set( get_dev_setting( 'make_excel_visible' ) )

        cb = Tk.Checkbutton( base_frame, text="Make Excel visible",
                                variable=self.make_excel_visible  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W )

        # Remaining Excel
        grid_row += 1
        self.keep_remaining_excel = Tk.IntVar()
        self.keep_remaining_excel.set( get_dev_setting( 'keep_remaining_excel' ) )
        cb = Tk.Checkbutton( base_frame, text="Keep remaining Excel instances instead of killing",
                                variable=self.keep_remaining_excel  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W )

        # DnD Debug
        grid_row += 1
        self.enable_dnd_debug = Tk.IntVar()
        self.enable_dnd_debug.set( get_dev_setting( 'enable_dnd_debug' ) )
        cb = Tk.Checkbutton( base_frame, text="enable DnD Debug",
                                variable=self.enable_dnd_debug  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W )

        # global grab cannot be set befor windows is 'viewable'
        # and this happen in mainloop after this function returns
        # Thus, it is needed to delay grab setting of an interval
        # long enough to make sure that the window has been made
        # 'viewable'
        if self.grab == 'global':
            self.after(100, self.grab_set_global )
        else:
            pass # local grab is set by parent class constructor

        self.update()

    def buttonbox( self ):
        box = Tk.Frame( self )
        box.pack( pady=10 )
        Dialog.buttonbox( self, frame=box )

    def clear_settings( self ):
        ok = MessageBox.askyesno( 'Warninig', 'This operations is not cancelable. OK?' )
        if ok:
            clear_settings()

    def apply( self ):  # overrides parent class method
        take_screenshots    = self.take_screenshots.get()
        screenshot_folder   = self.screenshot_folder.get()
        if ENABLE_HIRES_MODE_CHANGE:
            set_dev_setting( 'hiresolution',        self.hiresolution.get() )
        set_dev_setting( 'take_screenshots',    take_screenshots )
        set_dev_setting( 'screenshot_folder',   screenshot_folder )
        set_dev_setting( 'intensity_reduction', self.intensity_reduction.get() )
        set_dev_setting( 'reduction_cycle',     self.reduction_cycle.get() )
        set_dev_setting( 'reduction_start',     self.reduction_start.get() )
        set_dev_setting( 'no_usable_q_limit',   self.no_usable_q_limit.get() )
        set_dev_setting( 'log_memory_usage',    self.log_memory_usage.get() )
        set_dev_setting( 'log_xray_lpm_params', self.log_xray_lpm_params.get() )
        set_dev_setting( 'show_num_iterations', self.show_num_iterations.get() )
        set_dev_setting( 'suppress_defer_test', self.suppress_defer_test.get() )
        set_dev_setting( 'use_datgnom',         self.use_datgnom.get() )
        set_setting( 'enable_drift_simulation', self.enable_drift_simulation.get() )
        set_dev_setting( 'zx_add_constant',     self.zx_add_constant.get() )
        set_dev_setting( 'add_conc_const',      self.add_conc_const.get() )
        set_dev_setting( 'individual_bq_ingore', self.individual_bq_ingore.get() )
        set_dev_setting( 'make_excel_visible',   self.make_excel_visible.get() )
        set_dev_setting( 'keep_remaining_excel', self.keep_remaining_excel.get() )
        set_dev_setting( 'enable_dnd_debug',    self.enable_dnd_debug.get() )

        if take_screenshots:
            from OurScreenShot import screenshot
            self.parent.after( 500, lambda: screenshot( widget=self.parent, log=True ) )
        os.environ[ PYTHON_DEMO_MONITOR ]   = monitor = self.monitor.get()
        if monitor != self.monitor_previous:
            new_geometry = adjusted_geometry(self.parent.geometry())
            self.parent.geometry( new_geometry )
            print( 'changed geometry according to monitor selection', monitor, new_geometry )
        self.applied    = True



"""
    InitialInfo.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import os
import logging
import numpy as np
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from .OptimizerSettings import OptimizerSettings

class PeProxy:
    def __init__(self, optinit_info, optimizer, init_params):
        from molass_legacy.Peaks.PeakEditor import get_default_maxnum_trials
        self.previous_optimizer_folder = get_setting('optimizer_folder')
        self.previous_rg_folder = os.path.join(self.previous_optimizer_folder, 'rg-curve')
        self.sd = optinit_info.sd
        self.treat = optinit_info.treat
        self.n_components = optimizer.n_components
        self.fullopt = optimizer
        self.drift_type = optinit_info.drift_type
        self.init_params = init_params
        self.n_iterations = optinit_info.n_iterations
        self.maxnum_trials = get_default_maxnum_trials(self.n_components - 1)   # for consistency, see PeakEditor.get_n_components()
        self.param_init_type = optinit_info.param_init_type

    def load_settings(self):
        # creating logger here because this method will be called first
        # not in the __init__ which is too early.
        self.logger = logging.getLogger(__name__)
        settings = OptimizerSettings(param_init_type=self.param_init_type)
        settings.load(optimizer_folder=self.previous_optimizer_folder)
        self.logger.info("previous settings will be used through PeProxy.")

    def get_sd(self):
        return self.sd

    def get_treat(self):
        return self.treat

    def get_dsets(self):
        from .FullOptInput import FullOptInput
        # note that the analysis_folder should have been renewed by the time of this call

        """ this is required later in the show_peak_editor_impl.
            to clarify this necessity,
            find where the equivalent mkdirs is done in the non-proxy normal process
        """
        optimizer_folder = os.path.join(get_setting('analysis_folder'), 'optimized')
        if False:
            if not os.path.exists(optimizer_folder):
                from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry
                mkdirs_with_retry(optimizer_folder)

        if os.path.exists(self.previous_rg_folder):
            set_setting("trust_rg_curve_folder", True)

        set_setting("rg_curve_folder", self.previous_rg_folder)
        rg_folder = os.path.join(optimizer_folder, 'rg-curve')

        self.logger.info("optimizer_folder=%s", optimizer_folder)
        self.logger.info("previous_rg_folder=%s", self.previous_rg_folder)
        self.logger.info("new rg_folder=%s", rg_folder)

        # rg_folder will be expected to be copied from the previous_rg_folder

        sd = self.sd
        self.fullopt_input = FullOptInput(sd=sd, corrected_sd=None, rg_folder=rg_folder)
        self.dsets = self.fullopt_input.get_dsets(compute_rg=False)
        return self.dsets

    def get_drift_type(self):
        return self.drift_type

    def get_n_components(self):
        return self.n_components

    def get_seed(self):
        return np.random.randint(100000, 999999)

    def get_init_params(self):
        return self.init_params

    def get_function_class(self):
        class_ = self.fullopt.__class__
        return class_, class_.__name__

    def get_param_init_type(self):
        return self.param_init_type

    def get_n_iterations(self):
        return self.n_iterations

    def get_maxnum_trials(self):
        return self.maxnum_trials

class InitialInfo:
    def __init__(self, sd_copy=None, treat=None, pe=None, settings=None, result=None):

        if result is None:
            assert sd_copy is not None
            assert treat is not None
            assert pe is not None

            dsets = pe.get_dsets()
            dsets.relocate_rg_folder()

            drift_type = pe.get_drift_type()
            n_components = pe.get_n_components()
            n_iterations = pe.get_n_iterations()
            seed = pe.get_seed()
            init_params = pe.get_init_params()
            fullopt_class_not_used, class_code = pe.get_function_class()
            fullopt = pe.fullopt
            param_init_type = pe.get_param_init_type()
            maxnum_trials = pe.get_maxnum_trials()

            treat.save()

            settings = OptimizerSettings(param_init_type=param_init_type)
            settings.save()
 
            job_list = []
            composite = None
        else:
            # result
            sd_copy = result.sd_copy
            dsets = result.dsets
            drift_type = result.drift_type
            n_components = result.n_components
            n_iterations = result.n_iterations
            seed = result.seed
            init_params = result.init_params
            class_code = result.class_code
            fullopt = result.fullopt
            maxnum_trials = result.get_num_jobs()
            job_list = result.get_job_list()
            treat = result.treat
            settings = result.settings          # OptimizerSettings must be loaded before fullopt construction
                                                # which is done in FullOptResult

            param_init_type = settings.get("param_init_type")
            composite = result.composite

        self.treat = treat
        self.sd = sd_copy
        self.dsets = dsets
        self.drift_type = drift_type
        self.n_components = n_components
        self.init_params = init_params
        self.n_iterations = n_iterations
        self.seed = seed
        self.class_code = class_code            # str used in the optimizer
        self.fullopt = fullopt
        self.param_init_type = param_init_type
        self.bounds_type = settings.get("bounds_type")
        self.maxnum_trials = maxnum_trials
        self.job_list = job_list
        self.composite = composite
